package com.wincornixdorf.vnc.global;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.text.SimpleDateFormat;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.SimpleFormatter;

import org.apache.commons.lang3.StringUtils;

import com.wincornixdorf.vnc.rfb.conf.LoginParameters;
import com.wincornixdorf.vnc.rfb.conf.VNCSrvCfg;

/**
 * The LogToWebServiceAppender can be used as an ordinary logging appender
 * which will write LogEvents flagged as Level.FINER to a WebLogger. For that
 * reason one should NEVER log anything from in here using that log level as
 * one might accidentally create an infinite loop.
 * <p>
 * This class was developed from existing code which I took the liberty to
 * encapsulate in a decent class that adheres the most primitive of OO principles.
 * Some things it does seem downright wrong to me, but I don't even know what
 * it is this code DOES log to. Hence I have to trust the existing solution to the
 * bitter end and hope for the best. :)
 * <p>
 * One of the strangest things I noticed is the password one can supply on the
 * commandline is not used anywhere in the code.
 * <p>
 * I had to rewrite this class virtually speaking in the last minute in an effort
 * to move it from a Log4J based solution to a JavaLogging based one. This class
 * would probably best be derived from SocketHandler than from Handler, but I
 * didn't quite have the time to look into it.
 */
public class LogToWebServiceAppender extends Handler {
    /**
     * The format in which Date/Time Stamps will be written to the log
     */
    private static SimpleDateFormat formatter = new SimpleDateFormat("yyyyMMddHHmmss");

    /**
     * Contains hostname port and credentials required to access the web logger.
     */
    private LoginParameters logServer = null;

    /***/
    private VNCSrvCfg vncServer = null;

    /***/
    private URL loggingServerURL = null;

    /**
     * TODO: Check of this is REALLY necessary. I think the flush and close methods offered
     * by this class to satisfy the interface of the superclass should indeed be empty.
     */
    private Writer logWriter = null;

    private boolean failed2Publish = false;

    public LogToWebServiceAppender(LoginParameters logServerIn, VNCSrvCfg vncServerIn) {
        if (logServerIn != null) logServer = logServerIn;
        if (vncServerIn != null) vncServer = vncServerIn;

        setLevel(Level.INFO);
        setFormatter(new SimpleFormatter());

        loggingServerURL = logServer.getLogURL();
        Level lvl = VNCViewer.isDivine() ? Level.WARNING : Level.SEVERE;
        String msg = "The web logger is malconfigured! Opening connections will not be possible.";

        if (loggingServerURL != null) {
            lvl = Level.FINEST;
            msg = "Web Logger URL : [" + loggingServerURL + "]";
        }

        VNCViewer.pen.log(lvl, msg);
    }

    /**
     * This "security feature" of course is absolute rubbish. It's like setting
     * the fox to keep the geese. The idea is to only start a session, if the
     * application can write a message to the log indicating a session is
     * initiated. Since the log server can be specified on the cmdline, an
     * attacker trying to hide their tracks only has to have the log message
     * sent to a server of their own.
     *
     * @return False, if writing a log message to the log server failed.
     */
    public boolean isReadyToLog() {
        //    	return true;
        return loggingServerURL != null && !failed2Publish;
    }

    public void publish(LogRecord event) {
        // Just covering the basics here ...
        if (event == null || loggingServerURL == null) return;

        // May seem strange, but by convention only messages of level
        // FINER are being written to the Web Logger. The original code would log
        // everything, rendering the log almost unreadable. So don't be surprised
        // if you find messages that seem to get logged twice. The first one will
        // be written to the applications local log, the other to the web log.
        if (event.getLevel() != Level.FINER) return;

        String lm = "Internal error";

        boolean remoteSupport = vncServer.isRemoteSupport();

        try {
            //Formatter f = getFormatter();
            lm = event.getMessage();
            //lm = f.format( event);
            URLConnection conn = loggingServerURL.openConnection();

            conn.setDoInput(true);     // If I set this false, maybe I won't have to read the answer??!
            conn.setDoOutput(true);
            conn.setUseCaches(false);

            // I understand we need to set up a writer to write to whatever is listening
            // on the other end of the connection ...
            logWriter = new OutputStreamWriter(conn.getOutputStream());

            String un = logServer.getUsername(); //logUser
            String ds = formatter.format(new java.util.Date());
            String application = remoteSupport ? "Remote Support" : "Remote Control";

            String ch;
            if (remoteSupport && !StringUtils.isEmpty(vncServer.getDestIP())) {
                ch = new StringBuilder(vncServer.getDestIP()).append(" / ").append(vncServer.getHostParams().getSessionId()).toString();
            } else if (remoteSupport) {
                ch = vncServer.getHostParams().getSessionId();
            } else {
                ch = vncServer.getLabel();
            }

            // I had to learn the hard way the URLEncoder can't handle NULL arguments.
            // As the Applet will cease to work without a proper error message, in
            // the event of errors, I will sooner send nonsense than see it die.
            un = fallback(un, "unknown");
            ch = fallback(ch, "unknown");
            ds = fallback(ds, "now");
            lm = fallback(lm, "FAILED to format log message!");

            String le = URLEncoder.encode("user", "UTF-8") + "=" +
                    URLEncoder.encode(un, "UTF-8") + "&" +
                    URLEncoder.encode("datetime", "UTF-8") + "=" +
                    URLEncoder.encode(ds, "UTF-8") + "&" +
                    URLEncoder.encode("application", "UTF-8") + "=" +
                    URLEncoder.encode(application, "UTF-8") + "&" +
                    URLEncoder.encode("action", "UTF-8") + "=" +
                    URLEncoder.encode(lm, "UTF-8") + "&" +
                    URLEncoder.encode("object", "UTF-8") + "=" +
                    URLEncoder.encode(ch, "UTF-8");

            VNCViewer.pen.log(Level.FINEST, "WebLogger message: " + le);

            //System.err.println( "user     = {" + un + "}");
            //System.err.println( "datetime = {" + ds + "}");
            //System.err.println( "action   = {" + lm + "}");
            //System.err.println( "object   = {" + ch + "}");

            logWriter.write(le);
            logWriter.flush();

            // OMFG! One MUST read the answer from the logging servlet or the message
            // WON'T GET LOGGED! Mere curiosity but what if sending the REALLY IMPORTANT
            // error message is the last thing the application is able to do???!
            BufferedReader rd = new BufferedReader(new InputStreamReader(conn.getInputStream()));
            while (rd.readLine() != null) { /*Process line...*/}

            logWriter.close();
            rd.close();
        } catch (IOException e) {
            failed2Publish = true;
            System.err.println("Failed to write the following message to the web logger:\n[" + event.getMessage() + "]");
            if (!VNCViewer.isDivine())
                System.err.println("All connections will be closed as the web logger is out of the loop.");
        }
    }

    private String fallback(String s, String f) {
        return s != null ? s : f;
    }

    public void flush() {
        if (logWriter == null) return;

        try {
            logWriter.flush();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public void close() {
        if (logWriter == null) return;

        try {
            logWriter.close();
        } catch (IOException ioe) {
            ioe.printStackTrace();
        }
    }
}
