package com.wincornixdorf.vnc.global;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import javax.swing.Action;

public class VNCActionsPool {

	private HashMap<VNCAction.Goal, VNCAction> actions = new HashMap<VNCAction.Goal, VNCAction>();

    public VNCActionsPool( VNCViewer v) {
    	// Build all actions and sort them into a hashmap for fast lookup
    	for ( VNCAction.Goal goal : VNCAction.Goal.values())
    		actions.put( goal, new VNCAction( v, goal));

        // The "options" and "clipboard" actions are a toggle action of global
        // nature. By default they are always hidden when the app is started.
    	actions.get( VNCAction.Goal.OPTIONS).putValue( Action.SELECTED_KEY, new Boolean( false));
    	actions.get( VNCAction.Goal.CLIPBOARD).putValue( Action.SELECTED_KEY, new Boolean( false));
    	actions.get( VNCAction.Goal.TOGGLE_FULLSCREEN).putValue( Action.SELECTED_KEY, new Boolean( false));
    	actions.get( VNCAction.Goal.TOGGLE_CONN_STATUS_BAR).putValue( Action.SELECTED_KEY, new Boolean( false));
    	
        // The about action is always available
    	actions.get( VNCAction.Goal.ABOUT).setEnabled( true);
    }
    
    public VNCAction getActionByGoal( VNCAction.Goal g) {
    	return actions.get( g);
    }

    public void setEnabled( VNCAction.Goal goal, boolean b) {
    	VNCAction a = getActionByGoal( goal);
    	a.setEnabled( b);
    }

    public List<VNCAction> getActions( Set<VNCAction.Goal> goals) {
    	Vector<VNCAction> rval = new Vector<VNCAction>();
    	Iterator <VNCAction.Goal> it = goals.iterator();
    	while ( it.hasNext())
    		rval.add( actions.get( it.next()));
    	return rval;
    }

    /**Composes a list of actions, that are to be made globally available through
     * their keyboard shortcuts. They have to be declared specifically in the 
     * VNCSessionPoolPanel in order for they mnemonics to work as advertised.
     * @return A vector containing the VNCAction objects to add to the VNCSessionPoolPanel.*/
    public Vector<VNCAction> getGloballyAccessibleActions() {
        Vector<VNCAction> gaActions = new Vector<VNCAction>();
        
        VNCAction action = null;
        Iterator<VNCAction> itActions = actions.values().iterator();
        while ( itActions.hasNext()) {
        	action = itActions.next();
        	if ( action.getValue( Action.ACCELERATOR_KEY) != null)
        		gaActions.add( action);
        }
		
        return gaActions;    	
    }

    /**Composes the list of actions, that are to be made accessible through the glasspane.
     * @param applet Set true, if running this application as an Applet so actions that don't work
     * in an Applet can be excluded where required.
     * @return A vector containing {@link VNCAction} objects.*/
    public Vector<VNCAction> getMainGlasspaneActions( boolean applet) {
        Vector<VNCAction> gpactions = new Vector<VNCAction>();
        
        gpactions.add( applet ? actions.get( VNCAction.Goal.DELETE_SESSION) : actions.get( VNCAction.Goal.QUIT));
        gpactions.add( actions.get( VNCAction.Goal.SHOW_FB));
        gpactions.add( actions.get( VNCAction.Goal.SHOW_FT));
        gpactions.add( actions.get( VNCAction.Goal.SHOW_LG));

        gpactions.add( actions.get( VNCAction.Goal.OPTIONS));
        gpactions.add( actions.get( VNCAction.Goal.CLIPBOARD));

        gpactions.add( actions.get( VNCAction.Goal.CLOSE_FB_CONNECTION));
        gpactions.add( actions.get( VNCAction.Goal.RSPOWERUP));
        gpactions.add( actions.get( VNCAction.Goal.RSPOWERDOWN));
        gpactions.add( actions.get( VNCAction.Goal.CLOSE_FT_CONNECTION));
        gpactions.add( actions.get( VNCAction.Goal.DISABLE_CONTROLS));

        gpactions.add( actions.get( VNCAction.Goal.META));
        gpactions.add( actions.get( VNCAction.Goal.CTRLALTDEL));
        gpactions.add( actions.get( VNCAction.Goal.TOGGLE_CONN_STATUS_BAR));
        gpactions.add( actions.get( VNCAction.Goal.LIST_SESSIONS));
        
        gpactions.add( actions.get( VNCAction.Goal.SHOW_CONNECTION_DOCTOR));
        gpactions.add( actions.get( VNCAction.Goal.REFRESH));
        gpactions.add( actions.get( VNCAction.Goal.ABOUT));

        gpactions.add( actions.get( VNCViewer.isDivine() ? VNCAction.Goal.TOGGLE_FULLSCREEN : VNCAction.Goal.GLASS));
        
        return gpactions;
    }

    public Vector<VNCAction> getKeyRequestActions() {
        Vector<VNCAction> gpactions = new Vector<VNCAction>();

        gpactions.add( actions.get( VNCAction.Goal.META));
        gpactions.add( actions.get( VNCAction.Goal.CTRLALTDEL));
        gpactions.add( actions.get( VNCAction.Goal.BATTLA));
        gpactions.add( actions.get( VNCAction.Goal.ALTTAB));
        gpactions.add( actions.get( VNCAction.Goal.ALTF4));
        gpactions.add( actions.get( VNCAction.Goal.ALTGR_EUR));
        gpactions.add( actions.get( VNCAction.Goal.ALTGR_AT));
        gpactions.add( actions.get( VNCAction.Goal.QUICK_KEYS));
        
        return gpactions;
    }
}

