package com.wincornixdorf.vnc.gui;

//
//  Copyright (C) 2001 HorizonLive.com, Inc.  All Rights Reserved.
//  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
//
//  This is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This software is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this software; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
//  USA.
//

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.util.Observer;
import java.util.Observable;

import javax.swing.Box;
import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JTextArea;
import javax.swing.SwingUtilities;

import com.wincornixdorf.vnc.rfb.RFBConnJob;
import com.wincornixdorf.vnc.rfb.RFBHandler;
import com.wincornixdorf.vnc.rfb.RFBHandlerEvent;
import com.wincornixdorf.vnc.rfb.RFBHandlerEvent.Category;

public class ClipboardPanel extends JPanel
    implements ActionListener, Observer
{
    private static final long   serialVersionUID = 4038163623265824679L;
    
    private JTextArea           textArea        = null;
    //private JLabel              lWarning        = null; // Not used anymore ... might come in handy some day
    private JButton             clearButton     = null;
    private JButton             sendButton      = null;

    private RFBHandler          conn            = null;
    
    // Constructor.
    public ClipboardPanel() {
        super( new BorderLayout( 3, 3));

        JPanel jpTextAreaAndWarningLabel = new JPanel( new BorderLayout( 0, 0));
        textArea = new JTextArea( "");
        textArea.setPreferredSize( new Dimension( 400, 300));
        textArea.setEditable( true);
        textArea.setFont( new Font( "Monospaced", textArea.getFont().getStyle(), 14));

        //lWarning = new JLabel( "This could be YOUR warning String! Send me a word if you want 2 advertise here ;)");
        //lWarning.setAlignmentX( CENTER_ALIGNMENT);
        //lWarning.setForeground( Color.ORANGE);
        //lWarning.setVisible( false);

        jpTextAreaAndWarningLabel.add( textArea, BorderLayout.CENTER);
        //jpTextAreaAndWarningLabel.add( lWarning, BorderLayout.SOUTH);

        JPanel jpControls = new JPanel();

        FlowLayout l = new FlowLayout( FlowLayout.RIGHT);
        jpControls.setLayout( l);
        
        clearButton = new JButton( "Clear");
        clearButton.setMnemonic( KeyEvent.VK_L);
        clearButton.addActionListener( this);

        jpControls.add( Box.createRigidArea( new Dimension( 0, 3)));

        sendButton = new JButton( "Send");
        sendButton.setMnemonic( KeyEvent.VK_S);
        sendButton.addActionListener( this);

        jpControls.add( sendButton);
        jpControls.add( clearButton);
        
        add( jpTextAreaAndWarningLabel, BorderLayout.CENTER);
        add( jpControls, BorderLayout.SOUTH);
    }

    public void setConnection( RFBHandler c) {
        if ( c == null || conn != null) return;
        conn = c;
        conn.addObserver( this);
    }

    // Respond to button presses
    public void actionPerformed( ActionEvent evt) {
        if ( evt.getSource() == clearButton)
            textArea.setText( "");

        if ( evt.getSource() == sendButton)
            conn.queue( RFBConnJob.setClipboard( conn, textArea.getText()));
    }

    private void updateClipboard() {
        if ( conn == null) return;
        
        textArea.setText( conn.getClipboardText());
        if ( isVisible()) textArea.selectAll();
    }

    public void update( Observable o, Object arg) {
        if ( o != conn || !(arg instanceof RFBHandlerEvent)) return;
        
        RFBHandlerEvent e = (RFBHandlerEvent)arg;
        if ( e.operation != Category.CLIPCHANGED) return;

        // This is certainly overblown as setText is thread safe, but I am
        // not sure about the selectAll bit. Besides ... it's good practice.
        try {
             SwingUtilities.invokeLater( new Runnable() {
                 public void run() {
                     updateClipboard();
                 }
             });
        } catch ( Exception ex) { }
    }
}
