/**
 * 
 */
package com.wincornixdorf.vnc.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.Frame;
import java.awt.HeadlessException;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Date;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextArea;

import com.wincornixdorf.vnc.global.VNCViewer;
import com.wincornixdorf.vnc.global.VNCViewerOptions;
import com.wincornixdorf.vnc.rfb.RFBHandler;
import com.wincornixdorf.vnc.rfb.VNCSession;
import com.wincornixdorf.vnc.rfb.conf.LoginParameters;
import com.wincornixdorf.vnc.rfb.conf.VNCSrvCfg;
import com.wincornixdorf.vnc.rfb.exceptions.LoginException;

/**
 * @author dic
 *
 */
public class ConnectionDoctor extends JDialog implements ActionListener 
{
    private static final long serialVersionUID = 9075201605946993899L;

    private VNCSession  sess    = null;
    
    private JTextArea   jtaFB   = null;
    private JTextArea   jtaFT   = null;
    private JTextArea   jtaFBFT = null;

    private JButton     jbtnOk  = null;

    /**
     * @param f The {@link Frame} to put the new Widget into.
     * @param s The {@link VNCSession}, that is to be displayed
     * @param o The {@link VNCViewerOptions} used to configure the Session.
     */
    public ConnectionDoctor( Frame f, VNCSession s, VNCViewerOptions o)
    {
        super( f, "Connection Doctor", true);
        sess = s;

        setAlwaysOnTop( true);
        setDefaultCloseOperation( JFrame.DISPOSE_ON_CLOSE);
        setLayout( new BorderLayout());

        JTabbedPane jtp = new JTabbedPane();
        jtp.setPreferredSize( new Dimension( 1024, 768));

        jtaFBFT = buildTextComponent();
        jtaFB   = buildTextComponent();
        jtaFT   = buildTextComponent();

        String diagnoseFB = runDiagnosis( sess.getFB(), o);
        String diagnoseFT = runDiagnosis( sess.getFT(), o);
        String split      = "\n================================================================================\n\n";

        jtaFBFT.setText( diagnoseFB + split + diagnoseFT);
        jtaFB.setText( runDiagnosis( sess.getFB(), o));
        jtaFT.setText( runDiagnosis( sess.getFT(), o));

        jtp.add( "Combined"   ,     new JScrollPane( jtaFBFT));
        jtp.add( "Framebuffer",     new JScrollPane( jtaFB));
        jtp.add( "Filetransfer",    new JScrollPane( jtaFT));

        jtp.setMnemonicAt( 0, KeyEvent.VK_C);
        jtp.setMnemonicAt( 1, KeyEvent.VK_B);
        jtp.setMnemonicAt( 2, KeyEvent.VK_T);

        FlowLayout lojpButton = new FlowLayout( FlowLayout.RIGHT, 3, 3);
        JPanel jpButton = new JPanel( lojpButton);
        jbtnOk = new JButton( "Ok");
        jbtnOk.setMnemonic( KeyEvent.VK_O);
        jpButton.add( jbtnOk);
        jbtnOk.addActionListener( this);

        getContentPane().add( jtp, BorderLayout.CENTER);
        getContentPane().add( jpButton, BorderLayout.SOUTH);

        pack();
    }

    private JTextArea buildTextComponent() {
        JTextArea jta = new JTextArea( "Nothing to report here!");
        jta.setEditable( false);
        jta.setFont( new Font( Font.MONOSPACED, jta.getFont().getStyle(), 14));
        jta.setLineWrap( true);
        jta.setWrapStyleWord( true);
        return jta; 
    }

    private void print( VNCViewerOptions o, PrintWriter pw) {
        if ( pw == null) return;

        if ( o  == null) {
            pw.println( "WARNING! No options to print!");
            return;
        }

        Dimension    	d  = o.getCanvasDim();
        VNCSrvCfg    	sc = o.getVNCServerConnectionParameters();
        LoginParameters ls = o.getLogServerParameters();

        if ( sc != null) {
            String rep = sc.requiresRepeaterToConnect() ? sc.getHostnameAndPort( true, false) : ">>>DIRECT CONNECTION<<<";
            pw.println( String.format( "Pseudo URL        : [%1$s]", sc.getPseudoURL()));
            pw.println( String.format( " ↳ Connection     : [%1$s]", sc.getConMode()));
            pw.println( String.format( " ↳ Username       : [%1$s]", sc.getHostParams().getUsername()));
            pw.println( String.format( " ↳ Destination    : [%1$s]", sc.getHostnameAndPort( false, false)));
            pw.println( String.format( " ↳ Repeater       : [%1$s]", rep));
            pw.println( String.format( " ↳ Label          : [%1$s]", sc.getLabel()));
            pw.println( String.format( " ↳ Filename       : [%1$s]", sc.filename));
        }

        pw.println( String.format( "App. mode         : [%1$s]", o.getAppMode().toString()));
        pw.println( String.format( "Default Encoding  : [%1$s]", o.getDefaultEncoding().toString()));
        pw.println( String.format( "Keystore          : [%1$s]", o.getFilenameCryptstore( true)));
        pw.println( String.format( "Truststore        : [%1$s]", o.getFilenameCryptstore( false)));
        pw.println( String.format( "Canvas Dimension  : [%1$d x %2$d]", d.width, d.height));
        pw.println( String.format( "Blank Console     : [%1$b]", o.wantsConsoleBlanked()));
        pw.println( String.format( "Show Controls     : [%1$b]", o.wants2ShowControls()));
        pw.println( String.format( "Is in Applet Mode : [%1$s]", o.isRunAsApplet()));
        pw.println( String.format( "Has its own Frame : [%1$s]", o.isRunInSeparateFrame()));

        pw.println( String.format( "Repeater Message  : [%1$s]", o.getVNCServerConnectionParameters().getRepeaterMsg()));

        if ( ls != null) {
            pw.println( String.format( "Log Server        : [%1$s]", ls.getLogURL()));
            pw.println( String.format( "LogPath           : [%1$s]", ls.getLogPath()));
        }
    }
    
    private String runDiagnosis( RFBHandler h, VNCViewerOptions o) {

        StringWriter sw = new StringWriter();
        PrintWriter pw = new PrintWriter( sw, true);

        pw.println( "Bug Report generated on [" + new Date() + "]");
        pw.println( "by [" + VNCViewer.appBaseline( true) + "]");
        pw.println( "");
        pw.println( " [" + new Date() + "]");
        pw.println( "");
        pw.println( "Options passed into Viewer");
        pw.println( "==========================");
        print( o, pw);

        pw.println( "");
        pw.println( "VNC connection details");
        pw.println( "======================");
        if ( h != null) {
            VNCSrvCfg    sc = o.getVNCServerConnectionParameters();

            String cType = h.canHandleFileTransfers() ? "Filetransfer" : "Framebuffer";
            pw.println( String.format( "Connection Type   : [%1$s]", cType));
            pw.println( String.format( "Currenly in Phase : [%1$s]", h.getPhase().toString()));
            pw.println( String.format( "Server offered    : [%1$s]", sc.getProtocolVersion( true)));
            pw.println( String.format( "Client used       : [%1$s]", sc.getProtocolVersion( true)));

            Throwable cc = h.getCauseOfCrash();
            LoginException lie = getConnFailedException( cc);
            if ( lie != null) {
                if ( lie.hasConnectionFailedReason()) {
                    pw.println( "");
                    pw.println( "Connection Failed Message :");
                    pw.println( "===========================");
	                pw.println( lie.getConnectionFailedReason());
                } else {
                    pw.println( "");
                    pw.println( "Failed to log in          :");
                    pw.println( "===========================");
                    pw.println( lie.getMessage());
                }
            }

            pw.println( "");
            pw.println( "Stack Trace");
            pw.println( "===========");
            if ( cc != null) cc.printStackTrace( pw);
            else pw.println( ">>> No stack trace on record <<<");

        } else pw.println( ">>> No connection object available <<<");

        return sw.toString();
    }

    /**Searches the stack trace of the argument for the first occurrence of a
     * {@link ConnFailedException} and returns it where encountered.
     * @param t The {@link Throwable} to examine.
     * @return The first Cause of type {@link ConnFailedException} found in the
     * stack trace, or null if there is no such element.*/
    private LoginException getConnFailedException( Throwable t) {
    	if ( t == null) return null;
    	
    	LoginException lie = null;
    	
    	do {
    		if ( t instanceof LoginException) lie = (LoginException)t;
    		t = t.getCause();
    	} while ( t != null && lie == null);

    	return lie;
    }
    
    @Override
    public void actionPerformed( ActionEvent e) {
        if ( e.getSource() == jbtnOk) dispose();
    }

}
