package com.wincornixdorf.vnc.gui;

import java.util.HashMap;
import java.util.List;
import java.awt.*;
import java.awt.event.*;

import javax.swing.*;

import java.awt.image.BufferedImage;

import com.wincornixdorf.vnc.global.VNCAction;

/**Warning! This file is UTF-8 encoded and MUST NOT be edited in a non UTF-8
environment. Failure to do so may break the application. If the testcharacters
are broken check your editors encoding setting. If they are still garbled some
insensitive clod broke it :(
Testcharacters: Ä Ö Ü ß ä ö ü Խ Ф Щ д г 걎 겡 괱 궤 꿹 خ ج ث ؤ */

public class GlassPane extends JComponent
        implements MouseListener, MouseMotionListener
{
    private static final long	serialVersionUID	= 2479747306823443666L;
    public  static final String DEF_ACTIONSET   	= "DEFAULT";
    public  static final String KEYS_ACTIONSET   	= "QUICKKEYS";

    // Colors used for drawing the action buttons
    private static final Color  COLOUR_FILL    		= new Color( 141, 139, 230);
    private static final Color  COLOUR_BORDER    	= new Color(  60,  60, 155);
    
    // Used internally to indicate that no action is presently highlighted
    private static Point    ILLEGAL     = new Point( -1, -1);

    // Percent given to msg and ctrl bounds
    private static double   pc          = 0.1;

    // Percent Action icons arc
    private static double   pArcWidth   = 0.2;
    private static double   pArcHeight  = 0.2;

    // Percent Action icons "insets"
    private static double   pInset      = 0.2;

    // The Component the GlassPane will superimpose the icons upon
    private Component       c           = null;

    // number of rows and columns in which the icons will be displayed
    private int             rows       = 2;
    private int             cols       = 4;

    // The presently "highlighted" action
    private Point           highlighted = ILLEGAL;

    // Position of the mouse pointer by the last time the HL item changed.
    private Point           mouse       = ILLEGAL;

    private String			currAS		= DEF_ACTIONSET;
    private HashMap<String, List<VNCAction>> hmActs = new HashMap<String, List<VNCAction>>();

    private boolean         zoomAll     = false;

    public GlassPane( Component cIn) {
        c = cIn;

        addMouseListener( this);
        addMouseMotionListener( this);
    }

    protected String safelyGetAS( String name) {
    	if ( name == null || name.isEmpty()) return DEF_ACTIONSET;
    	return name;
    }

    public void addActionSet( String name, List<VNCAction> lActions) {
    	hmActs.put( safelyGetAS( name), lActions);
    }

    public void activateActionSet( String name, int colsIn, int rowsIn) {
    	currAS = safelyGetAS( name);
    	rows = rowsIn;
    	cols = colsIn;
    }

    protected List<VNCAction> getCurrentActionsSet() {
    	return hmActs.get( currAS);
    }

    protected VNCAction getAction( int row, int col) {
        int pos = row * cols + col;
        List<VNCAction> actions = getCurrentActionsSet();
        if ( actions == null || pos < 0 || pos >= actions.size()) return null;

        VNCAction a = null;

        try {
            a = actions.get( pos);
        } catch ( ArrayIndexOutOfBoundsException e) {
            e.printStackTrace();
        }
        return a;
    }

    protected double dist( Point p1, Point p2) {
        double a = p1.x - p2.x;
        double b = p1.y - p2.y;
        return Math.sqrt( a * a + b * b); // Rejoice! It's Pythagoras ;)
    }

    protected void paintAction( Graphics g, Rectangle r, Action a, double aw, double ah) {
        boolean toggled = isSelected( a);
        g.setColor( toggled ? COLOUR_BORDER : COLOUR_FILL);
        g.fillRoundRect( r.x, r.y, r.width, r.height, (int)(r.width * aw), (int)(r.height * ah));
        g.setColor( toggled ? COLOUR_FILL : COLOUR_BORDER);
        g.drawRoundRect( r.x, r.y, r.width, r.height, (int)(r.width * aw), (int)(r.height * ah));

        if ( a == null) return;

        Rectangle ri   = shrink( r, 0.9);
        ImageIcon ii   = (ImageIcon)(a.getValue( VNCAction.GLASSPANE_ICON));

        if ( ii == null) return;

        Image     iimg = ii.getImage();

        if ( !a.isEnabled()) {
            int iimgW = iimg.getWidth( this);
            int iimgH = iimg.getHeight( this);

            BufferedImage bi = new BufferedImage( iimgW, iimgH, BufferedImage.TYPE_BYTE_GRAY);
            Graphics gi = bi.getGraphics();
            float fa = (float)1.0;
            float fr = (float)0.552941176471;
            float fg = (float)0.545098039216;
            float fb = (float)0.901960784314;

            Color trans = new Color( fr, fg, fb, fa);
            gi.drawImage( iimg, 0, 0, trans, null);
            gi.dispose();

            Graphics2D g2d = (Graphics2D) g;
            g2d.drawImage( bi, ri.x, ri.y, ri.width, ri.height, (JComponent)this);
        } else {
            g.drawImage( iimg, ri.x, ri.y, ri.width, ri.height, (JComponent)this);
        }
    }

    /**Displays the highlighted actions name and description into the top
     * and the bottom message area respectively.
     * @param g The {@link Graphics} context to paint into.*/
    protected void drawActionLabel( Graphics g) {
    	VNCAction hlAct = getAction( highlighted.y, highlighted.x);
        if ( hlAct == null) return;

        String name = (String)hlAct.getValue( Action.NAME);
        String desc = (String)hlAct.getValue( Action.SHORT_DESCRIPTION);

        Color bg = Color.BLACK;
        Color fg = Color.WHITE;
        
        SharedGraphicsUtils.printDoubleBold( g, name, msgBounds( true), g.getFont(), bg, fg);
        SharedGraphicsUtils.printDoubleBold( g, desc, msgBounds( false), g.getFont(), bg, fg);
        

        String moco = VNCAction.getDefaultKeyStrokeAsString( hlAct.getGoal());
        if ( moco.isEmpty()) return;
        
    	fg = hlAct.isEnabled() ? Color.GREEN : Color.DARK_GRAY;
        SharedGraphicsUtils.printDoubleBold( g, moco, shortcutBounds(), g.getFont(), bg, fg);
    }

    private void paintRect( Graphics g, Rectangle r) {
        g.drawRect( r.x, r.y, r.width, r.height);
    }
   
    /**Draws a set of boxes on the screen. This method is serves debugging purposes only.
     * @param g The {@link Graphics} context to paint into.*/
    @SuppressWarnings("unused")
    private void paintDbgBoxes( Graphics g) {
        g.setColor( Color.red);
        paintRect( g, centerBounds());

        g.setColor( Color.green);
        paintRect( g, ctrlBounds( true));

        g.setColor( Color.blue);
        paintRect( g, ctrlBounds( false));

        g.setColor( Color.green);
        paintRect( g, msgBounds( true));

        g.setColor( Color.blue);
        paintRect( g, msgBounds( false));

        g.setColor( Color.black);
        paintRect( g, centerSquaredBounds());
    }

    /**This method will draw all actions to the given Graphics context. If no
     * action is selected, all actions will be displayed having the same size.
     * Otherwise all actions will be drawn zoomed by a factor which depends on
     * how far the actions icon is away from the current mouse cursors position.
     * @param g The {@link Graphics} context to paint into.*/
    protected void paintZoomingAll( Graphics g) {
        Rectangle rCSB = centerSquaredBounds();

        if ( highlighted != ILLEGAL) {
            for ( int r = 0; r < rows; r++) {
                for ( int c = 0; c < cols; c++) {
                    Rectangle ico = iconBounds( r, c);
                    double dMouse2Icon = dist( mouse, calcCenter( ico));
                    double z =  pInset * ( dMouse2Icon / rCSB.width);
                    Rectangle iconZoomed = iconBounds( r, c, z);
                    paintAction( g, iconZoomed, getAction( r, c), pArcWidth, pArcHeight);
                }
            }
        } else {
            for ( int r = 0; r < rows; r++)
                for ( int c = 0; c < cols; c++)
                    paintAction( g, iconBounds( r, c, pInset), getAction( r, c), pArcWidth, pArcHeight);
        }

        drawActionLabel( g);
    }

    protected void paintZoomingHighlightedOnly( Graphics g) {
        for ( int r = 0; r < rows; r++)
            for ( int c = 0; c < cols; c++)
                paintAction( g, iconBounds( r, c, pInset), getAction( r, c), pArcWidth, pArcHeight);

        if ( highlighted == ILLEGAL) return;

        Rectangle rCSB = centerSquaredBounds();
        Rectangle rHLIco = iconBounds( highlighted.y, highlighted.x);
        double dMouse2HLIconCenter = dist( mouse, calcCenter( rHLIco));
        double z =  pInset * ( dMouse2HLIconCenter / rCSB.width);
        Rectangle rHLIconZoomed = iconBounds( highlighted.y, highlighted.x, z);
        paintAction( g, rHLIconZoomed, getAction( highlighted.y, highlighted.x), pArcWidth, pArcHeight);

        drawActionLabel( g);
    }

    protected void paintComponent( Graphics g) {
        if ( zoomAll) paintZoomingAll( g);
        else paintZoomingHighlightedOnly( g);
    }

    /**@return The {@link Rectangle} enclosing the Component the glasspane gets
     * superimposed upon. In other words: This is the space available to draw into.*/
    protected Rectangle maxBounds() {
        return c.getBounds();
    }

    /**@return The {@link Rectangle} the action tiles could get painted into. The widget
     * however will try to get a square area to paint into, which will be obtained
     * through centerSquaredBounds.*/
    protected Rectangle centerBounds() {
        Rectangle rm = maxBounds();
        double lw = rm.width * pc;
        double lh = rm.height * pc;

        return r( rm.x + lw, rm.y + lh, rm.width - 2 * lw, rm.height - 2 * lh);
    }

    /**@return A square shaped {@link Rectangle} centered within the one by
     * {@linkplain centerBounds} sized to fit into the "logical parent".*/
    protected Rectangle centerSquaredBounds() {
        Rectangle rm = centerBounds();
        Rectangle wi = new Rectangle( rm);

        int m = rm.height < rm.width ? rm.height : rm.width;
        rm.setSize( m, m);

        return center( rm, wi);
    }

    /**@param left You may guess three times. If you need a third attempt ... Back away from the keyboard.
     * @return The {@link Rectangle} to the left or right of the centerSquaredBounds.*/
    protected Rectangle ctrlBounds( boolean left) {
        Rectangle rm = maxBounds();
        double lw = rm.width * pc;
        double lh = rm.height * pc;

        if ( left) return r( rm.x, rm.y + lh, lw, rm.height - 2 * lh);
        else return r( rm.x + rm.width - lw, rm.y + lh, lw, rm.height - 2 * lh);
    }

    private Rectangle shortcutBounds() {
    	Rectangle theLot         = centerSquaredBounds();
    	Point     theLotsCenter  = calcCenter( theLot);
    	double    shortcutWidth  = theLot.width * 0.8;
    	double    shortcutHeight = theLot.height * 0.2;
    	
    	return r( theLotsCenter.x - shortcutWidth / 2, theLotsCenter.y - shortcutHeight / 2,  shortcutWidth, shortcutHeight);
    }

    protected Rectangle msgBounds( boolean top) {
        Rectangle rm = maxBounds();
        double lh = rm.height * pc;

        if ( top) return r( rm.x, rm.y, rm.width, lh);
        else return r( rm.x, rm.y + rm.height - lh, rm.width, lh);
    }
    
    protected Rectangle iconBounds( int row, int col) {
        Rectangle rm = centerSquaredBounds();
        double lw = rm.width / cols;
        double lh = rm.height / rows;

        return r( rm.x + col * lw, rm.y + row * lh, lw, lh);
    }

    protected Rectangle iconBounds( int row, int col, double sf) {
        Rectangle rm = centerSquaredBounds();
        double lw = rm.width / cols;
        double lh = rm.height / rows;

        double iw = lw * sf;
        double ih = lh * sf;

        return r( rm.x + col * lw + iw, rm.y + row * lh + ih, lw - 2 * iw, lh - 2 * ih);
    }

    protected Rectangle shrink( Rectangle r, double f) {
        Rectangle rNyu = new Rectangle( r);
        rNyu.setSize( (int)(r.width * f), (int)(r.height * f));
        return center( rNyu, r);
    }

    protected Rectangle r( double x, double y, double w, double h) {
        return new Rectangle( (int)x, (int)y, (int)w, (int)h);
    }

    protected Rectangle center( Rectangle r, Rectangle within) {
        double cxWithin = within.x + within.width / 2;
        double cyWithin = within.y + within.height / 2;
        double rwm      = r.width / 2;
        double rhm      = r.height / 2;

        return new Rectangle( (int)(cxWithin - rwm), (int)(cyWithin - rhm), r.width, r.height);
    }

    protected Point calcCenter( Rectangle r) {
        return new Point( (int)(r.x + r.width / 2), (int)(r.y + r.height / 2));
    }

    protected Point actionAt( Point pMouse) {
        Rectangle r = centerSquaredBounds();
        if ( !r.contains( pMouse)) return ILLEGAL;

        pMouse.translate( -1 * r.x, -1 * r.y);

        double tw = r.width / cols;
        double th = r.height / rows;

        int hlX = (int)( pMouse.x / tw);
        int hlY = (int)( pMouse.y / th);

        return new Point( hlX, hlY);
    }

    protected boolean hlChanged( Point old, Point nyu) {
        if ( old == null || nyu == null) return false;
        return old.x != nyu.x || old.y != nyu.y;
    }

    public void setHL( Point p) {
        if ( p == null) return;

        if ( p != ILLEGAL) {
            int corrX = p.x;
            int corrY = p.y;

            if ( corrX < 0) corrX = 0;
            if ( corrY < 0) corrY = 0;

            if ( corrX >= cols) corrX = cols - 1;
            if ( corrY >= rows) corrY = rows - 1;
            p.setLocation( corrX, corrY);
        }

        highlighted = p;
    }

    public void mouseDragged( MouseEvent e) {
    }

    public void mouseMoved( MouseEvent e) {
        //Point oldHL = hl;

        setHL( actionAt( e.getPoint()));
        mouse = e.getPoint();

        //if ( !hlChanged( oldHL, hl)) return;

        /*if ( hl != notHL)*/ repaint();
    }

    public void mouseClicked( MouseEvent e) {
        //System.out.println( "HLx: " + hl.x + " HLy: " + hl.y);
        Rectangle r = iconBounds( highlighted.y, highlighted.x);
        if ( !r.contains( e.getPoint())) return;

        Action a = getAction( highlighted.y, highlighted.x);
        if ( a != null && a.isEnabled()) {
            // This code deals with toggle actions :/
            if ( isToggle( a)) a.putValue( Action.SELECTED_KEY, !isSelected( a));

            a.actionPerformed( new ActionEvent( this, highlighted.x * highlighted.y, ""));
            setVisible( false);
        }
    }

    public void mouseEntered( MouseEvent e) {
    }

    public void mouseExited( MouseEvent e) {
    }

    public void mousePressed( MouseEvent e) {
    }

    public void mouseReleased( MouseEvent e) {
    }

    private boolean isToggle( Action a) {
        if ( a == null) return false;
        Boolean b = (Boolean)a.getValue( Action.SELECTED_KEY);
        if ( b == null) return false;
        else return true;
    }

    private boolean isSelected( Action a) {
        if ( !isToggle( a)) return false;
        Boolean b = (Boolean)a.getValue( Action.SELECTED_KEY);
        return b.booleanValue();
    }
}

