/**
 *
 */
package com.wincornixdorf.vnc.gui;


import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.LayoutManager;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.font.FontRenderContext;
import java.awt.font.LineBreakMeasurer;
import java.awt.font.TextAttribute;
import java.awt.font.TextLayout;
import java.awt.geom.Rectangle2D;
import java.text.AttributedCharacterIterator;
import java.text.AttributedString;

import javax.swing.ImageIcon;
import javax.swing.JPanel;

import com.wincornixdorf.vnc.global.VNCAction;
import com.wincornixdorf.vnc.global.VNCViewer;
import com.wincornixdorf.vnc.rfb.RFBHandler;
import com.wincornixdorf.vnc.rfb.exceptions.LoginException;

/**
 * @author dic
 */
public class RFBHandlerJPanel extends JPanel implements MouseListener {
    /**
     *
     */
    private static final long serialVersionUID = 552907996102650703L;


    private final ImageIcon iDisconnect = new ImageIcon(this.getClass().getResource("/chardevice.png"));
    private final ImageIcon iConnecting = new ImageIcon(this.getClass().getResource("/64x64/connect_creating.png"));
    private final ImageIcon iNotPlugged = new ImageIcon(this.getClass().getResource("/64x64/connect_no.png"));
    private final ImageIcon iConPlugged = new ImageIcon(this.getClass().getResource("/64x64/connect_established.png"));
    private final ImageIcon iConRunning = new ImageIcon(this.getClass().getResource("/64x64/display.png"));
    private final ImageIcon iConClosing = new ImageIcon(this.getClass().getResource("/64x64/desktop_stop.png"));
    private final ImageIcon iConnFailed = new ImageIcon(this.getClass().getResource("/64x64/bug.png"));

    protected RFBHandler hdlr = null;
    protected Image dblBuff = null;
    protected boolean neverHadAConnection = true;
    protected boolean drawDC = true;

    // Pretend we have the mouse 2 prevent display of gray scaled RD on startup
    protected boolean hasMouse = true;


    /**
     *
     */
    public RFBHandlerJPanel() {
        super();
    }

    public RFBHandlerJPanel(LayoutManager loMgr) {
        super(loMgr);
    }

    public void setConnection(RFBHandler c) {
        if (hdlr != null || hdlr == c || c == null) return;

        hdlr = c;
        neverHadAConnection = false;
    }

    protected boolean isDisconnected() {
        return hdlr == null || hdlr.getPhase() != RFBHandler.Phase.PROCESSING;
    }

    protected void spawnDblBuffer(int width, int height) {
        if (dblBuff != null &&
                dblBuff.getWidth(this) == width &&
                dblBuff.getHeight(this) == height)
            return;

        width = width <= 0 ? 30 : width;
        height = height <= 0 ? 30 : height;

        dblBuff = createImage(width, height);
    }

    /**
     * Will generate the message to be seen in the event the connection is no
     * longer running. The method will look at a set of status flags and other
     * variables to establish the likely cause of the disconnect and generate a
     * message which will hopefully help the user to understand why the connection
     * went down and if possible how to fix it again.
     *
     * @return The aforementioned message.
     */
    protected String getStatusMessage() {
        if (hdlr == null) return neverHadAConnection ? "Connecting ..." : "No connection object set.";

        RFBHandler.Phase phase = hdlr.getPhase();

        switch (phase) {
            case PENDING:
                if (hdlr.hasAnErrorMessage())
                    return "The connection has failed. Please try again. " + hdlr.getHRErrorMessage();
                else
                    return "This connection is waiting to be opened.";
            case PREPARING:
                return "Setting up socket.";
            case CONNECTING:
                return "Opening connection with the server.";
            case ROUTING:
                return "Negotiating with a repeater.";
            case AUTHENTICATING:
                return "Sending username and password to server.";
            case NEGOTIATING:
                return "Negotiating pixel format and encoding with the server.";
            case DISCONNECTING:
                return "Closing down sockets, preparing connection for reuse.";
            default:
                return null;
        }
    }

    private ImageIcon getStatusIcon() {
        if (hdlr == null) return iNotPlugged;

        RFBHandler.Phase phase = hdlr.getPhase();

        switch (phase) {
            case PENDING:
                if (hdlr.hasAnErrorMessage()) return iDisconnect;
                else return iNotPlugged;
            case PREPARING:
                return iConnecting;
            case CONNECTING:
                return iConnecting;
            case ROUTING:
                return iConPlugged;
            case AUTHENTICATING:
                return iConPlugged;
            case NEGOTIATING:
                return iConRunning;
            case DISCONNECTING:
                return iConClosing;
            default:
                return iDisconnect;
        }
    }

    private void drawImageInto(Graphics g, Image i, Rectangle2D r) {
        double min = Math.min(r.getWidth(), r.getHeight());
        g.drawImage(i, (int) (r.getCenterX() - min / 2), (int) (r.getCenterY() - min / 2), (int) min, (int) min, null);
    }


    @SuppressWarnings("unused")
    private void drawRect(Graphics g, Rectangle2D r2d, Color c) {
        Rectangle r = r2d.getBounds();
        g.setColor(c);
        g.drawRect(r.x, r.y, r.width, r.height);
    }

    @SuppressWarnings("unused")
    private void foo(Graphics2D g2d, Rectangle2D r, String s, Font font, Color color) {
        Rectangle rh = r.getBounds();
        AttributedString as = new AttributedString(s);
        as.addAttribute(TextAttribute.FONT, font);
        as.addAttribute(TextAttribute.FOREGROUND, color);

        // Get iterator for string
        AttributedCharacterIterator itChar = as.getIterator();
        // Get font context from graphics
        FontRenderContext rc = g2d.getFontRenderContext();
        // Create measurer
        LineBreakMeasurer lbreak = new LineBreakMeasurer(itChar, rc);

        int x = rh.x;
        int y = rh.y;

        while (lbreak.getPosition() < itChar.getEndIndex()) {
            TextLayout textLayout = lbreak.nextLayout(rh.width);
            y += textLayout.getAscent();
            textLayout.draw(g2d, x, y);
            y += textLayout.getDescent() + textLayout.getLeading();
        }
    }

    private void strikeOut(Graphics g, int gpw, int gph, int dx, int dy, Color c) {
        g.setColor(Color.black);
        g.fillRect(0, 0, gpw, gph);

        g.setColor(c);
        for (int x = 0; x < gpw; x += 2 * dx) {
            for (int y = 0; y < gph; y += 2 * dy) {
                g.fillRect(x, y, dx, dy);
                g.fillRect(x + dx, y + dy, dx, dy);
            }
        }
    }

    protected void showGlasspaneKeyTip(Graphics g) {
        boolean tipDisabled = hdlr != null && hdlr.getSettingsGB().hideGlasspaneTip;
        if (hasMouse || VNCViewer.isGlasspaneVisible() || tipDisabled) return;
        Dimension s = getSize();
        Rectangle2D r = new Rectangle2D.Double(10, 10, s.getWidth(), s.getHeight() * 0.8);
        String gpKS = VNCAction.getDefaultKeyStrokeAsString(VNCAction.Goal.GLASS);
        String qkKS = VNCAction.getDefaultKeyStrokeAsString(VNCAction.Goal.QUICK_KEYS);
        String msg = "To access the controls press [" + gpKS + "] or [" + qkKS + "] while this message is shown.";
        SharedGraphicsUtils.printDoubleBold(g, msg, r, g.getFont(), Color.black, Color.GREEN);
    }

    protected void showDisableControlsMessage(Graphics g) {
        if (VNCViewer.isGlasspaneVisible() || hdlr == null || !hdlr.isDisabledRemoteControlling()) return;
        Dimension s = getSize();
        Rectangle2D r = new Rectangle2D.Double(10, 10, s.getWidth(), s.getHeight() * 0.8);
        String msg = "Controls are disabled. Press Disable/Enable controls button to enable them.";
        SharedGraphicsUtils.printDoubleBold(g, msg, r, g.getFont(), Color.black, Color.GREEN);
    }

    protected void showDisconnected(Graphics g, Dimension d2showIn, String statusMsg) {
        if (statusMsg == null) return;

        Rectangle rc = new Rectangle(d2showIn);
        String url = "UNCONFIGURED";
        Image img = getStatusIcon().getImage();

        boolean connHasError = false;
        boolean showConnFailedErrorSymbol = false;

        if (hdlr != null) {
            url = hdlr.getServerParams().getPseudoURL();
            connHasError = hdlr.hasAnErrorMessage();

            showConnFailedErrorSymbol = (hdlr.getCauseOfCrash() instanceof LoginException) &&
                    ((LoginException) hdlr.getCauseOfCrash()).hasConnectionFailedReason();
        }

        strikeOut(g, (int) getWidth(), (int) getHeight(), 5, 5, Color.darkGray);

        double ratioUrl = 1.0 / 4.0;
        double ratioImg = 2.0 / 4.0;
        double ratioMsg = 1.0 / 4.0;
        double xMargin = rc.width / 10.0;
        double yMargin = rc.height / 10.0;

        Rectangle2D rAll = new Rectangle2D.Double(xMargin, yMargin, rc.width - 2 * xMargin, rc.height - 2 * yMargin);

        Rectangle2D rURL = new Rectangle2D.Double(rAll.getX(), rAll.getY(),
                rAll.getWidth(), rAll.getHeight() * ratioUrl);

        Rectangle2D rIco = new Rectangle2D.Double(rAll.getX(), rURL.getY() + rURL.getHeight(),
                rAll.getWidth(), rAll.getHeight() * ratioImg);

        Rectangle2D rMsg = new Rectangle2D.Double(rAll.getX(), rIco.getY() + rIco.getHeight(),
                rAll.getWidth(), rAll.getHeight() * ratioMsg);

        Rectangle2D rVer = new Rectangle2D.Double(rc.getX(), rc.getMaxY() - yMargin,
                rc.width - (xMargin / 4), yMargin);

        Color cURL = Color.white;
        Color cIco = Color.lightGray;
        Color cMsg = connHasError ? Color.red : Color.white;

        //drawRect( g, rAll, Color.cyan);
        //drawRect( g, rIcon, Color.green);
        //drawRect( g, rMess, Color.magenta);

        // Show the URL as prominently as possible ... we all know every1 loves 2 send screenshots :(
        SharedGraphicsUtils.printDoubleBold(g, url, rURL, g.getFont(), Color.black, cURL);

        // The Message goes @ the bottom.
        SharedGraphicsUtils.printDoubleBold(g, statusMsg, rMsg, g.getFont(), Color.black, cMsg);

        // The unplugged screen is eye-candy :)
        drawImageInto(g, img, rIco);

        // If a "connection failed" message is available, we also show a cute little bug superimposed
        // on the unplugged screen to indicate to the user they should consult the connection doctor.
        if (showConnFailedErrorSymbol) {
            double bugW = rIco.getWidth() / 4.0;
            double bugH = rIco.getHeight() / 4.0;
            double offsetX = rIco.getWidth() / 2.0 - bugW / 2.0;
            double offsetY = rIco.getHeight() / 2.0 - bugH / 0.8;

            Rectangle2D rCFIcon = new Rectangle2D.Double(rIco.getX() + offsetX, rIco.getY() + offsetY, bugW, bugH);

            drawImageInto(g, iConnFailed.getImage(), rCFIcon);
        }

        g.setColor(Color.green);
        SharedGraphicsUtils.drawStringInto(g, VNCViewer.appBaseline(true), rVer, g.getFont(), false,
                Component.RIGHT_ALIGNMENT, Component.BOTTOM_ALIGNMENT);

        // The viewer offers a human readable and a developer error message users can switch
        // in between. Since most people don't know that, it must be shown here!
        // Since we want that information only when showing an error message, the next line will
        // bounce if it is one of the startup info messages.
        if (!connHasError) return;

        if (VNCViewer.isDivine()) {
            // I swear if I get another screen-dump reading "General Network Error" I'll go postal!
            SharedGraphicsUtils.printDoubleBold(g, VNCViewer.ASK_THE_DOCTOR, rIco, g.getFont(), Color.black, cIco);
        } else {
            SharedGraphicsUtils.printDoubleBold(g, VNCViewer.DONT_ASK_THE_DOCTOR, rIco, g.getFont(), Color.black, cIco);
        }

        showGlasspaneKeyTip(g);
    }

    protected boolean isLegal(Dimension d) {
        return (d != null && d.height * d.width != 0);
    }

    @Override
    public void mouseEntered(MouseEvent evt) {
        hasMouse = true;
    }

    @Override
    public void mouseExited(MouseEvent evt) {
        hasMouse = false;
    }

    @Override
    public void mouseClicked(MouseEvent arg0) {
        // TODO Auto-generated method stub
    }

    @Override
    public void mousePressed(MouseEvent arg0) {
        // TODO Auto-generated method stub
    }

    @Override
    public void mouseReleased(MouseEvent arg0) {
        // TODO Auto-generated method stub
    }

//    @Override
//    public void setPreferredSize( Dimension preferredSize) {
//    	VNCViewer.pen.severe( ">>>>>>>>>>>>>>>>>>>>>> Got a hunch [" + preferredSize + "]");
//    	super.setPreferredSize( preferredSize);
//    	revalidate();
//    }

//    @Override
//    public void setPreferredSize( Dimension preferredSize) {
//    	VNCViewer.pen.severe( ">>>>>>>>>>>>>>>>>>>>>> Got a hunch [" + preferredSize + "]");
//    	if ( !isLegal( preferredSize))
//    		preferredSize = hdlr != null ? hdlr.getRemoteDesktopOriginalSize() : null;
//    		
//    	if ( !isLegal( preferredSize))
//    		preferredSize = new Dimension( 1280, 720);
//    	
//    	super.setPreferredSize( preferredSize);
//    }

    @Override
    protected void paintComponent(Graphics g) {
        if (!drawDC) {
            super.paintComponent(g);
            return;
        }

        spawnDblBuffer(getWidth(), getHeight());
        Graphics gDblBuff = dblBuff.getGraphics();
        paintInternal(gDblBuff);
        g.drawImage(dblBuff, 0, 0, this);
    }

    protected void paintInternal(Graphics g) {
        if (isDisconnected())
            showDisconnected(g, getSize(), getStatusMessage());
    }
}
