package com.wincornixdorf.vnc.gui;

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.Component;
import java.awt.Font;
import java.util.logging.SimpleFormatter;

import javax.swing.Action;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

import com.wincornixdorf.vnc.rfb.VNCSession;
import com.wincornixdorf.vnc.rfb.RFBHandler;
import com.wincornixdorf.vnc.rfb.conf.VNCSessionCfg;
import com.wincornixdorf.vnc.global.TextAreaAppender;
import com.wincornixdorf.vnc.global.VNCViewer;
import com.wincornixdorf.vnc.gui.ft.FTWidget;

public class VNCSessionPanel extends JPanel {

    private static final long serialVersionUID = 8980593635137796061L;

    private class VNCSessionTab extends JPanel {
        private static final long serialVersionUID = 935227635301044022L;
        public  JComponent          main    = null;
        public  JComponent          opts    = null;
        public  JComponent          clip    = null;

        public  ConcealablePanel    cpOpts  = null;
        public  ConcealablePanel    cpClip  = null;

        public VNCSessionTab( JComponent mainIn, JComponent optsIn, JComponent clipIn) {
            super( new BorderLayout());

            main = mainIn;
            opts = optsIn;
            clip = clipIn;

            if ( opts != null) {
                Icon iiOptions   = new ImageIcon( this.getClass().getResource( "/16x16/configure.png"));
                cpOpts = new ConcealablePanel( "Options", iiOptions, opts);
                add( cpOpts, BorderLayout.NORTH);
            }

            if ( main != null) add( main, BorderLayout.CENTER);

            if ( clip != null) {
                Icon iiClipboard = new ImageIcon( this.getClass().getResource( "/16x16/editpaste.png"));
                cpClip = new ConcealablePanel( "Clipboard contents", iiClipboard, clip);
                add( cpClip, BorderLayout.SOUTH);
            }
        }
    }

    private final static String TAB_FB = "FB";
    private final static String TAB_FT = "FT";
    private final static String TAB_LG = "LG";

//    private static int          blx             = 5;
//    private static int          bly             = 5;

//    private Boolean             fbVisible       = new Boolean( true);
//    private Boolean             ftVisible       = new Boolean( false);
//    private Boolean             lgVisible       = new Boolean( false);

    private VNCSession          session         = null;
    private TextAreaAppender    taa             = null;

    private VNCSessionTab       stFB            = null;
    private VNCSessionTab       stFT            = null;
    //private VNCSessionTab       stLG            = null;

    public  RemoteDesktopPanel  fbWidget        = null;
    private FTWidget            ftWidget        = null;
    private JTextArea           lgWidget        = null;
    public  SessionStatusPanel  statusPanel     = null;

    private JPanel              views           = null;

    private String              currentTab      = null;

    public VNCSessionPanel() {
        super( new BorderLayout());

        // Create the Widget for displaying the RFB
        JScrollPane spFB = new JScrollPane();
        fbWidget = new RemoteDesktopPanel( false);
        spFB.setViewportView( fbWidget);

        stFB = new VNCSessionTab( spFB, new JOptionsWidget( true), new ClipboardPanel());

        // Create the Widget for doing FT operations
        ftWidget = new FTWidget();
        stFT = new VNCSessionTab( ftWidget, new JOptionsWidget( false), null);

        // Create the Widget to show the log in
        lgWidget = new JTextArea();
        lgWidget.setEditable( false);
        lgWidget.setFont( new Font( "Monospaced", lgWidget.getFont().getStyle(), 14));

        JScrollPane spLG = new JScrollPane( lgWidget);
        spLG.setViewportView( lgWidget);

        // Place all three into a CardLayout controlled container
        views = new JPanel( new CardLayout());
        views.add( stFB, TAB_FB);
        views.add( stFT, TAB_FT);
        views.add( spLG, TAB_LG);

        statusPanel = new SessionStatusPanel();

        add( views, BorderLayout.CENTER);
        add( statusPanel, BorderLayout.SOUTH);

        showOnTop( TAB_FB);
    }

    public void setOptionsVisible( boolean b) {
        stFB.cpOpts.setVisible( b);
        stFT.cpOpts.setVisible( b);
    }

    public void setClipboardsVisible( boolean b) {
        stFB.cpClip.setVisible( b);
    }

    /**Puts the widget showing the remote desktop on display. This method must
     * be invoked on the EDT.
     * @return True, if successful.*/
    public boolean showFB() { return showOnTop( TAB_FB); }

    /**Puts the widget showing the remote filesystem on display. This method
     * must be invoked on the EDT.
     * @return True, if successful.*/
    public boolean showFT() { return showOnTop( TAB_FT); }

    /**Puts the widget showing the log on display. This method must be invoked
     * on the EDT.
     * @return True, if successful.*/
    public boolean showLG() { return showOnTop( TAB_LG); }

    public boolean isShowingFB() { return currentTab == TAB_FB; }
    public boolean isShowingFT() { return currentTab == TAB_FT; }
    public boolean isShowingLG() { return currentTab == TAB_LG; }

    public Component getInputComponent() { return fbWidget != null ? fbWidget : ftWidget; }

    private boolean showOnTop( final String s) {
        if ( s == null || s == currentTab) return false;
        CardLayout l = (CardLayout)(views.getLayout());
        l.show( views, s);
        currentTab = s;
        return true;
    }

    public void toggleStatusBar() {
    	statusPanel.setVisible( !statusPanel.isVisible());
    }

    public VNCSession getSession() {
        return session;
    }

    /**Set the panel to show the RFB / Files of the VNCConnection objects of the argument.
     * The session will be started automatically.
     * @param s The {@link VNCSession} to show. */
    public void setSession( VNCSession s) {
        if ( session != null || s == null) return;

        this.session = s;
        statusPanel.setSession( s);

        VNCSessionCfg cfg = session.getConfiguration();
        ((JOptionsWidget)(stFB.opts)).setOptions( cfg);
        ((JOptionsWidget)(stFT.opts)).setOptions( cfg);

        // attach an appender to the sessions log object
        // which will relay all messages to the logs TextArea
        taa = new TextAreaAppender( lgWidget);
        taa.setFormatter( new SimpleFormatter());
        VNCViewer.pen.addHandler( taa);

        if ( fbWidget != null) {
            RFBHandler connFB = session.getFB();
            fbWidget.setConnection( connFB);
            ((ClipboardPanel)(stFB.clip)).setConnection( connFB);
        }

        if ( ftWidget != null) {
            RFBHandler connFT = session.getFT();
            ftWidget.setConnection( connFT);
        }

        // ensure the appropriate widget is on top
        if ( cfg.cn.isFTOnly()) showFT(); else showFB();

        validate();
    }

    public void setSelectedKeys( Action aFB, Action aFT, Action aLG) {
        aFB.putValue( Action.SELECTED_KEY, new Boolean( isShowingFB()));
        aFT.putValue( Action.SELECTED_KEY, new Boolean( isShowingFT()));
        aLG.putValue( Action.SELECTED_KEY, new Boolean( isShowingLG()));
    }

    /**Extracts the current Preferences and puts them into the argument.
     * This solution is rather ugly as it has a quirk. The Framebuffer
     * preferences in the object passed out again will be overwritten completely!
     * @param cfgSess The {@link VNCSessionCfg} object to write the FT session configuration into.*/
    public void extract( VNCSessionCfg cfgSess) {
        if ( cfgSess == null) return;
        if ( ftWidget != null) ftWidget.extract( cfgSess.ft);
    }
}
