package com.wincornixdorf.vnc.gui.ft;

// Copyright (C) 2002-2005 Ultr@VNC Team.  All Rights Reserved.
// Copyright (C) 2004 Kenn Min Chong, John Witchel.  All Rights Reserved.
//
//This is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This software is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this software; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
//USA.
//

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.File;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Observable;
import java.util.Observer;
import java.util.Vector;
import java.util.logging.Level;

import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.RowSorter;
import javax.swing.SortOrder;
import javax.swing.SwingUtilities;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;
import javax.swing.table.TableModel;
import javax.swing.table.TableRowSorter;

import com.wincornixdorf.vnc.global.VNCViewer;
import com.wincornixdorf.vnc.gui.FSTableCellRenderer;
import com.wincornixdorf.vnc.gui.RFBHandlerJPanel;
import com.wincornixdorf.vnc.rfb.RFBConnJob;
import com.wincornixdorf.vnc.rfb.RFBHandler;
import com.wincornixdorf.vnc.rfb.RFBHandlerEvent;
import com.wincornixdorf.vnc.rfb.conf.SettingsFT;
import com.wincornixdorf.vnc.rfb.ft.FTFile;
import com.wincornixdorf.vnc.rfb.ft.FTPath;
import com.wincornixdorf.vnc.rfb.ft.RemoteFileSystem;

/**
 * @author John Witchel, Kenn Min Chong
 */

public class FTWidget extends RFBHandlerJPanel
        implements ActionListener, MouseListener, ListSelectionListener,
        Observer {
    private static final long serialVersionUID = -6473158207180231512L;

    private static Dimension preferredDims4ProgyBar = new Dimension(200, 20);
    private static Dimension preferredDims = new Dimension(800, 800);

    private static String DEFABORTMSG = "There are no file transfers running. Nothing was aborted";
    private static String FTWARN = "File Transfer Warning";
    private static String FTERRR = "File Transfer Error";
    private static String REAL_FT_VIEW = "REAL";
    private static String FAKE_FT_VIEW = "FAKE";

    private static int blx = 5;
    private static int bly = 5;

    private SettingsFT ftPrefs = null;

    private JPanel jpClientSide = null;
    private JPanel jpServerSide = null;
    private JPanel jpProgBarEtc = null;

    private FTCmdButtons cmdBtnsClient = null;
    private FTCmdButtons cmdBtnsServer = null;

    private JButton stopButton = null;

    private JTable clientFileView = null;

    private JTable serverFileView = null;
    private JPanel jpCardLOSrvView = null;
    private RFBHandlerJPanel jpHdlrStatus = null;

    private JTextField clientStatus = null;
    private JTextField serverStatus = null;

    private JProgressBar jProgressBar = null;

    /**
     * Contains the files and directories to be shown in the client FS View.
     * The Table used for displaying the FS requires the TableModel provided by
     * FTFileVector to work properly.
     */
    private RemoteFileSystem lfs = null;

    /**
     * This string will be added to the history on pressing the stop button.
     */
    private String abortErrorMsg = "";

    /**
     * This flag reflects whether the underlying FTConnection is busy or not.
     */
    private boolean hasBusyFTLine = false;

    /**
     * Workaround 4 a bug that can lead to a loss of the users preferences.
     * Do not mess with it!
     */
    private boolean mayExtractPreferences = false;

    public static String getFileDirCountString(int files, int dirs) {
        int total = files + dirs;
        return "Files: [" + files + "] Directories: [" + dirs + "] Total: [" + total + "]";
    }

    public FTWidget() {
        super();
        drawDC = false;
        createGUI();
    }

    @Override
    public void setConnection(RFBHandler h) {
        super.setConnection(h);
        if (hdlr != null) hdlr.addObserver(this);

        jpHdlrStatus.setConnection(h);

        // The Widget will observe the display options as the Date time
        // format to be used when displaying files is being set there!
        if (hdlr != null) {
            ftPrefs = hdlr.getSettingsFT();
            ftPrefs.addObserver(this);
        }
    }

    /**
     * Write the present FT settings into {@link SettingsFT} object.
     *
     * @param ftPrefs the {@link SettingsFT} object to write to.
     */
    public void extract(SettingsFT ftPrefs) {
        // The mayExtractPreferences parameter must be consulted 2 prevent
        // loosing the users preferences. As can be seen below, the bookmarks
        // are taken from the bookmark widget, which gets initialized when
        // the connection is coming up. If the connection does NOT come up for
        // any reason, this method would grab an EMPTY bookmark list, which
        // would then get persisted, overwriting the users actual list!
        if (ftPrefs == null || !mayExtractPreferences) return;
        ftPrefs.setPreferredClientDir(cmdBtnsClient.bookmark.getCurrentLocation());
        ftPrefs.setPreferredServerDir(cmdBtnsServer.bookmark.getCurrentLocation());
        ftPrefs.setClientBookmarks(cmdBtnsClient.bookmark.getBookmarks());
        ftPrefs.setServerBookmarks(cmdBtnsServer.bookmark.getBookmarks());
    }

    /**
     * Used internally to decide whether the download or upload button should be enabled or not.
     * This method is expecting the object presently selected in one of the views. It will return
     * true if the list contains items which can be transferred to the recipient.
     * <p>
     * If it is used in remote support mode, all directories are evaluated as non transferable
     */
    private boolean canBeTransferred(Object[] o) {
        if (o.length == 0) return false;

        int i = 0;
        boolean failed = false;
        //boolean remoteSupport = hdlr.getServerParams().isRemoteSupport();
        FTFile f;

        while (i < o.length && !failed) {
            f = ((FTFile) o[i++]);
            failed = !f.canBeTransferred();
            /*
            Disable possibility to transfer directory for remote support. Directory transfer is not supported in some versions of remote support.
            if (remoteSupport && f.isDirectory()) {
                return false;
            }*/
        }

        return !failed;
    }

    /**
     * Used internally to decide whether the download or upload button should be enabled or not.
     * This method is expecting the object presently selected in one of the views. It will return
     * true if the list contains items which can be transferred to the recipient.
     */
    private boolean canBeDeleted(Object[] o) {
        if (o.length == 0) return false;

        int i = 0;
        boolean failed = false;
        FTFile f = null;

        while (i < o.length && !failed) {
            f = ((FTFile) o[i++]);
            failed = !f.canBeDeleted();
        }

        return !failed;
    }

    private Object[] getSelectedEntriesIn(JTable t) {
        Vector<Object> vos = new Vector<Object>();

        // better safe than sorry ;)
        if (t == null || !(t == clientFileView || t == serverFileView))
            return vos.toArray();

        int[] sRows = t.getSelectedRows();

        for (int i = 0; i < sRows.length; i++) {
            int im = t.convertRowIndexToModel(sRows[i]);
            TableModel m = t.getModel();
            FTFile f = (FTFile) m.getValueAt(im, 0);
            vos.add(f);
        }

        return vos.toArray();
    }

    /**
     * Looks at the entries presently selected in the filesystem views and en-/disables
     * Buttons for those operations which can('t) be performed on these items. If \a client is
     * set, it will do so for the files selected in the local FS view. Otherwise it will look
     * at those in the remote FS view.
     */
    private void invalidateFTButtons(boolean client) {
        Object[] entries = getSelectedEntriesIn(client ? clientFileView : serverFileView);
        RemoteFileSystem v = client ? lfs : hdlr != null ? hdlr.getRFS() : null;
        boolean mint = v != null ? v.isInMintCondition() : false;
        boolean alive = false;
        //boolean dead = connection != null ? connection.isDeadInTheWater() : true;
        if (hdlr != null) alive = (!hasBusyFTLine && !hdlr.isDeadInTheWater());

        if (client) {
            cmdBtnsClient.btnDelete.setEnabled(canBeDeleted(entries));
            cmdBtnsClient.btnNewFolder.setEnabled(!mint && getClientCD().canWrite());
            cmdBtnsClient.btnRoot.setEnabled(getClientCD().getParentFile() != null);
            cmdBtnsClient.btnReload.setEnabled(getClientCD().canRead());
            cmdBtnsClient.btnTransfer.setEnabled(((alive && !hasBusyFTLine) && (canBeTransferred(entries))));
        } else {
            cmdBtnsServer.btnDelete.setEnabled(alive && canBeDeleted(entries));
            cmdBtnsServer.btnNewFolder.setEnabled(!mint && alive);
            cmdBtnsServer.btnRoot.setEnabled(alive);
            cmdBtnsServer.btnReload.setEnabled(alive);
            cmdBtnsServer.btnTransfer.setEnabled(((alive && !hasBusyFTLine) && (canBeTransferred(entries))));
            serverFileView.setEnabled(alive);
        }
    }

    /**
     * Convenience method for its parameterized counterpart, invoking it once for the
     * client and once for the server.
     */
    private void invalidateFTButtons() {
        invalidateFTButtons(true);
        invalidateFTButtons(false);
    }

    /**
     * Used internally to generate various error and warning messages.
     */
    private String getOnSrv(boolean onServer) {
        return getNLParam("on", onServer ? hdlr.getHost() : "localhost");
    }

    /**
     * Used internally to generate various error and warning messages.
     */
    private String getNLParam(String a, String b) {
        return "<br><i>" + a + "</i> <font color=blue>" + b + "</font>";
    }

    /**
     * Used internally to generate various error and warning messages.
     */
    private String markUp(String a) {
        return "<html>" + a + "</html>";
    }

    private void doNewFolder(boolean onServer) {
        Icon iNewFolder = new ImageIcon(this.getClass().getResource("/16x16/folder_new.png"));
        String currSrvDir = hdlr != null ? hdlr.getCurrentDir() : "Error: No connection!";
        String currCltDir = getClientCD().getAbsolutePath();
        String inFolder = onServer ? currSrvDir : currCltDir;
        String dlgCaption = "Create new directory";
        String lblCaption = "Enter name of new directory to <font color=green>create</font>" +
                getNLParam("in", inFolder) + getOnSrv(onServer);
        String name = (String) JOptionPane.showInputDialog(this, markUp(lblCaption), dlgCaption, JOptionPane.QUESTION_MESSAGE, iNewFolder, null, null);

        // if name is null, the user clicked "cancel" -> do NOT create folder
        if (name == null) return;

        if (onServer && hdlr != null)
            hdlr.queue(RFBConnJob.createDirectory(hdlr, hdlr.getCurrentDir() + name));
        else {
            File f = new File(getClientCD().getAbsolutePath() + File.separatorChar + name);
            f.mkdir();
            reloadCurrentClientDir();
            VNCViewer.pen.log(Level.CONFIG, "Created Local Directory [" + name + "]");
        }
    }

    private void doDelete(boolean onServer) {
        if (hdlr == null) return;

        if (onServer) {
            List<FTFile> selectedFiles = getSelectedFileInServerView();

            List<FTFile> filesToDelete = new ArrayList<FTFile>();
            StringBuilder confirmMessage = new StringBuilder("Are you sure you want to <font color=red>delete</font>");
            for (FTFile selectedFile : selectedFiles) {
                if (selectedFile == null || selectedFile.servesAsPlaceholder()) {
                    continue;
                }
                String sFileName = selectedFile.getFilename();

                if (selectedFile.isDirectory()) {
                    VNCViewer.pen.log(Level.WARNING, "The server does not support directory deletion.");
                    //connection.deleteRemoteDirectory( sFileName);
                    continue;
                }

                String fileName = hdlr.getCurrentDir() + sFileName;

                if (hdlr.getRFS().knowsFile(sFileName)) {

                    confirmMessage.append(getNLParam("file", fileName));
                    confirmMessage.append(getOnSrv(onServer));

                    filesToDelete.add(selectedFile);

                } else {
                    VNCViewer.pen.log(Level.WARNING, "Can't delete [" + sFileName + "]. File does not exist on server!");
                }
            }

            confirmMessage.append("?");
            if (!showConfirmDialog(confirmMessage.toString())) {
                return;
            }

            for (FTFile ftFile : filesToDelete) {
                hdlr.queue(RFBConnJob.deleteFile(hdlr, hdlr.getCurrentDir() + ftFile.getFilename()));
            }

        } else {
            List<FTFile> selectedFiles = getSelectedFileInClientView();

            StringBuilder confirmMessage = new StringBuilder("Are you sure you want to <font color=red>delete</font>");
            List<FTFile> filesToDelete = new ArrayList<FTFile>();
            for (FTFile selectedFile : selectedFiles) {
                if (selectedFile == null || selectedFile.servesAsPlaceholder() || selectedFile.isParent()) {
                    continue;
                }

                String sFileName = selectedFile.getFilename();
                File sFile = new File(getClientCD().getAbsolutePath() + File.separatorChar + sFileName);

                if (sFile.exists()) {

                    String type = selectedFile.isDirectory() ? "directory and its content" : "file";
                    confirmMessage.append(getNLParam(type, sFile.getAbsolutePath()));

                    filesToDelete.add(selectedFile);
                }
            }

            confirmMessage.append("?");

            if (!showConfirmDialog(confirmMessage.toString())) {
                return;
            }

            for (FTFile ftFile : filesToDelete) {
                deleteLocally(ftFile.getTheFile());
            }
            reloadCurrentClientDir();

        }
    }

    private void deleteLocally(File f) {
        if (f.isDirectory()) {
            File[] dirList = f.listFiles();
            if (dirList != null) {
                for (int i = 0; i < dirList.length; i++)
                    deleteLocally(dirList[i]);
                VNCViewer.pen.log(Level.CONFIG, "Deleted local directory [" + f.getAbsolutePath() + "]");
            }
        } else
            VNCViewer.pen.log(Level.CONFIG, "Deleted file [" + f.getAbsolutePath() + "]");
        f.delete();
    }

    private void downloadSelectedFileOrDirectory() {
        if (hdlr == null) return;

        List<FTFile> selectedFiles = getSelectedFileInServerView();
        for (FTFile selectedFile : selectedFiles) {
            if (selectedFile == null || selectedFile.servesAsPlaceholder()) {
                continue;
            }

            String sFileName = selectedFile.getFilename();

            String serverFileName = hdlr.getCurrentDir() + sFileName;
            String clientFileName = getClientCD().getAbsolutePath() + File.separatorChar + sFileName;
            File file2Write2 = new File(getClientCD(), sFileName);

            boolean confirmed = true;
            if (file2Write2.exists()) {
                String msg = "The local directory" + getOnSrv(false) +
                        "<br>already contains a file or folder" + getNLParam("named", sFileName) +
                        "<br><br>Are you sure you want to <font color=red>overwrite</font> it ?";

                confirmed = showConfirmDialog(msg);
            }
            if (selectedFile.isDirectory() && confirmed) {
                hdlr.queue(RFBConnJob.transfer(hdlr, sFileName, getClientCD().getAbsolutePath(), true, true));
            } else if (confirmed) {
                hdlr.queue(RFBConnJob.transfer(hdlr, serverFileName, clientFileName, true, false));
            }

        }

    }

    private void uploadSelectedFileOrDirectory() {
        if (hdlr == null) return;

        List<FTFile> selectedFiles = getSelectedFileInClientView();

        for (FTFile selectedFile : selectedFiles) {
            if (selectedFile == null || selectedFile.servesAsPlaceholder()) {
                continue;
            }
            String sFileName = selectedFile.getFilename();
            File file2Upload = new File(getClientCD(), sFileName);

            boolean confirmed = true;
            if (hdlr.getRFS().knowsFile(sFileName) ||
                    hdlr.getRFS().knowsDirectory(sFileName)) {
                String msg = "The remote directory" + getOnSrv(true) +
                        "<br>already contains a file or folder" + getNLParam("named", sFileName) +
                        "<br><br>Are you sure you want to <font color=red>overwrite</font> it ?";

                confirmed = showConfirmDialog(msg);
            }
            if (file2Upload.isDirectory() && confirmed) {
                hdlr.queue(RFBConnJob.transfer(hdlr, file2Upload.toString(), null, false, true));
            } else if (confirmed) {
                hdlr.queue(RFBConnJob.transfer(hdlr, file2Upload.getAbsolutePath(), hdlr.getCurrentDir(), false, false));
            }
        }

    }


    private boolean showConfirmDialog(String message) {
        int result = JOptionPane.showConfirmDialog(null, markUp(message), FTWARN, JOptionPane.YES_NO_OPTION);

        if (result == JOptionPane.CLOSED_OPTION || result == JOptionPane.NO_OPTION) {
            return false;
        }

        return true;
    }

    /**
     * Signals the connection to abort the present Filetransfer. Please keep in mind this will
     * only set a flag set in the connection object, which will be checked periodically. What the
     * connection will make out of it, is something else.
     */
    private void doStop() {
        if (hdlr == null) return;

        VNCViewer.pen.log(Level.WARNING, abortErrorMsg);
        hdlr.tryToAbortFileTransfer();
    }

    /**
     * Changes the local directory to the root node of the present drive. On *nix systems
     * the current directory will be set '/' on windumb boxes, the current directory will
     * become the topmost directory on the current drive ( e.g. "C:\Temp\foo" -> "C:\").
     */
    private void gotoRoot(boolean client) {
        FTCmdButtons cmdb = client ? cmdBtnsClient : cmdBtnsServer;
        String root = cmdb.bookmark.getRootOfCurrentPath();
        if (!root.isEmpty()) {
            if (client) changeLocalDirectory(new File(root));
            else hdlr.queue(RFBConnJob.changeDir(hdlr, root, true));
        }
    }

    /**
     * Pretty self-explanatory. If a directory is currently highlighted, we
     * will change to it. If it is a parent folder entry (aka "..") we will chdir there.
     */
    private void changeToDirectorySelectedInClientFileView() {
        List<FTFile> selectedDirectory = getSelectedFileInClientView();
        if (selectedDirectory.isEmpty() || selectedDirectory.size() > 1) return;

        FTFile file = selectedDirectory.get(0);
        if (file.servesAsPlaceholder()) {
            reloadCurrentClientDir();
            return;
        }

        File currentSelection = new File(getClientCD(), file.getFilename()); // Selection

        if (file.isParent()) { // The [..] entry is selected
            currentSelection = getClientCD().getParentFile();
            if (currentSelection != null) {
                changeLocalDirectory(currentSelection);
            } else {
                clientStatus.setText("You are at the root !");
            }
        } else if (currentSelection.isDirectory()) {
            changeLocalDirectory(currentSelection);
        }
    }

    /**
     * Tries to make the directory currently highlighted in the Server side
     * view the current directory, listing its contents in the Server side view.
     * The method will do nothing if no directory is presently selected.
     */
    private void changeToDirectorySelectedInServerFileView() {
        List<FTFile> selectedDirectory = getSelectedFileInServerView();
        if (hdlr == null || selectedDirectory.isEmpty() || selectedDirectory.size() > 1) {
            return;
        }

        FTFile currentSelection = selectedDirectory.get(0);
        if (currentSelection.servesAsPlaceholder()) {
            reloadCurrentServerDir();
            return;
        }

        if (currentSelection.isDirectory()) {
            hdlr.queue(RFBConnJob.changeDir(hdlr, currentSelection.getFilename(), false));
        }
    }

    /**
     * Issues a request to have the contents of the current directory reloaded
     * and their view updated accordingly.
     */
    private void reloadCurrentServerDir() {
        if (hdlr == null) return;

        // If the "drives" haven't been listed yet, do it NOW!
        if (cmdBtnsServer.bookmark.countSolidPaths() <= 0) {
            hdlr.queue(RFBConnJob.listDrives(hdlr));
            return;
        }

        hdlr.queue(RFBConnJob.changeDir(hdlr, cmdBtnsServer.bookmark.getCurrentLocation(), true));
    }

    /**
     * Looks for a suitable directory to use as the current local directory.
     * The method is guaranteed to to return an object.
     */
    private File getFallbackDirectory() {
        File[] r = File.listRoots();
        if (r == null || r.length <= 0 || r[0] == null)
            return new File(weAreHome() ? "/" : "C:\\"); // TODO: EPIC FAIL on my part :(

        return r[0];
    }

    /**
     * This method will return a File reference to the directory presently
     * shown in the client view. The method is guaranteed to return an object. If
     * the object is valid or not however is a different matter.
     */
    private File getClientCD() {
        String s = cmdBtnsClient.bookmark.getCurrentLocation();
        if (s == null || s.isEmpty()) return getFallbackDirectory();
        return new File(s);
    }

    private void changeLocalDirectory(File dir) {
        // Except 4 the fallback dir we reload only if the cd got changed.
        if (dir == null) return;

        String em = "Could not change to [" + dir + "]: ";
        if (!dir.isDirectory()) {
            clientStatus.setText(em + "Not a directory!");
            return;
        }
        if (!dir.canRead()) {
            clientStatus.setText(em + "Can't read directory!");
            return;
        }

        String lc = dir.getAbsolutePath();
        if (!lc.endsWith("" + File.separatorChar)) lc += File.separatorChar;
        cmdBtnsClient.bookmark.visitPath(lc);

        reloadCurrentClientDir();
    }

    private void reloadCurrentClientDir() {
        File[] dirList = getClientCD().listFiles();
        lfs.clear();

        if (dirList == null) {
            String m = "Internal error: Can't list [" + getClientCD() + "]";
            VNCViewer.pen.log(Level.WARNING, m);
            clientStatus.setText(m);
            getFTTableDM(true).forceUpdate();
            return;
        }

        int dirs = 0, files = 0;

        for (int i = 0; i < dirList.length; i++) {
            lfs.inSort(new FTFile(dirList[i]));
            if (dirList[i].isDirectory()) dirs++;
            if (dirList[i].isFile()) files++;
        }

        if (getClientCD().getParent() != null)
            lfs.inSort(new FTFile("..", false, false, false, 0, null, null, null));

        clientStatus.setText(getFileDirCountString(files, dirs));
        getFTTableDM(true).forceUpdate();
    }

    /**
     * Returns true if the local box has a *nix like FS and false if not.
     */
    private boolean weAreHome() {
        // TODO: This is sad and pathetic, but 4 now i don't know any better :(
        return File.separator.equalsIgnoreCase("/");
    }

    /**
     * This method initializes clientDrivesComboBox
     *
     * @return JComboBox
     */
    private void updateClientDrivesComboBox() {
        // Read in Drive letters from local disk
        File[] roots = File.listRoots();
        ArrayList<FTPath> al = new ArrayList<FTPath>();
        for (int i = 0; i < roots.length; i++)
            al.add(new FTPath(roots[i].toString()));

        cmdBtnsClient.bookmark.addSolidPaths(al);
    }

    private List<FTFile> getSelectedFileInClientView() {
        int[] selectedRows = clientFileView.getSelectedRows();
        if (selectedRows.length == 0) {
            return Collections.EMPTY_LIST;
        }

        List<FTFile> selectedFiles = new ArrayList<FTFile>();
        for (int selectedRow : selectedRows) {
            int modelIndex = clientFileView.convertRowIndexToModel(selectedRow);
            selectedFiles.add((FTFile) getFTTableDM(true).getValueAt(modelIndex, 0));
        }
        return selectedFiles;
    }

    private List<FTFile> getSelectedFileInServerView() {
        if (hdlr == null) {
            return Collections.EMPTY_LIST;
        }
        int[] selectedRows = serverFileView.getSelectedRows();
        if (selectedRows.length == 0) {
            return Collections.emptyList();
        }

        List<FTFile> selectedFiles = new ArrayList<FTFile>();
        for (int selectedRow : selectedRows) {
            int modelIndex = serverFileView.convertRowIndexToModel(selectedRow);
            selectedFiles.add(((FTFile) getFTTableDM(false).getValueAt(modelIndex, 0)));
        }

        return selectedFiles;
    }

    private void onDrvListChanged() {
        if (hdlr == null) return;

        // Sort the drives known to exist on the server into the combobox.
        // The action listener has to be removed temporarily to prevent the
        // contents of the remote directory from being reloaded on starting the
        // application as this may be quite time consuming.
        cmdBtnsServer.bookmark.removeActionListener(this);
        cmdBtnsServer.bookmark.addSolidPaths(hdlr.getKnownServerDrives());
        cmdBtnsServer.bookmark.addActionListener(this);

        // TODO: It may prove useful to clear the remote directory list as the drive we are looking at may have been unmounted.
        updateRemoteFileSystemSummary();
    }

    private void updateRemoteFileSystemSummary() {
        // Ask the bookmark widget to display the new current directory.
        // To prevent the widget from triggering another change directory
        // request, we have keep this widget from receiving Action events
        // for the time being.
        cmdBtnsServer.bookmark.removeActionListener(this);
        cmdBtnsServer.bookmark.visitPath(hdlr.getCurrentDir());
        cmdBtnsServer.bookmark.addActionListener(this);

        serverStatus.setText(getFileDirCountString(hdlr.getFilesCount(), hdlr.getDirectoriesCount()));
    }

    private void createGUI() {
        abortErrorMsg = DEFABORTMSG;

        // ------------------------
        // assemble the Client side
        // ------------------------
        jpClientSide = new JPanel(new BorderLayout(blx, bly));
        cmdBtnsClient = new FTCmdButtons(false, "Local");
        cmdBtnsClient.btnTransfer.setEnabled(false);
        cmdBtnsClient.addListener(this);

        lfs = new RemoteFileSystem("local filesystem");
        clientFileView = new JTable(new FTTableDataModel(lfs));
        setPreferredColumnSizes(true);
        clientFileView.setDefaultRenderer(lfs.getColumnClass(0), new FSTableCellRenderer());
        clientFileView.getSelectionModel().setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);

        clientFileView.addMouseListener(this);
        clientFileView.getSelectionModel().addListSelectionListener(this);
        clientFileView.setRowHeight(32);
        clientFileView.setShowHorizontalLines(false);
        clientFileView.setShowVerticalLines(false);

        TableRowSorter<FTTableDataModel> trsClt = new TableRowSorter<FTTableDataModel>(getFTTableDM(true));
        List<RowSorter.SortKey> sortKeys = new ArrayList<RowSorter.SortKey>();
        sortKeys.add(new RowSorter.SortKey(0, SortOrder.ASCENDING));
        trsClt.setSortKeys(sortKeys);
        trsClt.setComparator(0, FTComparator.getFilenameComparator());
        trsClt.setComparator(1, FTComparator.getSizeComparator());
        trsClt.setComparator(2, FTComparator.getDateComparator());
        trsClt.setComparator(3, FTComparator.getDateComparator());
        trsClt.setComparator(4, FTComparator.getDateComparator());
        clientFileView.setRowSorter(trsClt);
        trsClt.sort();

        JScrollPane jspc = new JScrollPane();
        jspc.setViewportView(clientFileView);

        clientStatus = new JTextField("");
        clientStatus.setEditable(false);

        jpClientSide.add(cmdBtnsClient, BorderLayout.NORTH);
        jpClientSide.add(jspc, BorderLayout.CENTER);
        jpClientSide.add(clientStatus, BorderLayout.SOUTH);

        // ------------------------
        // assemble the Server side
        // ------------------------
        jpServerSide = new JPanel(new BorderLayout(blx, bly));
        cmdBtnsServer = new FTCmdButtons(true, "Not connected");
        cmdBtnsServer.setFTBtnsEnabled(false);
        cmdBtnsServer.addListener(this);

        serverFileView = new JTable();
        serverFileView.getSelectionModel().setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
        serverFileView.addMouseListener(this);
        serverFileView.getSelectionModel().addListSelectionListener(this);
        serverFileView.setRowHeight(32);
        serverFileView.setShowHorizontalLines(false);
        serverFileView.setShowVerticalLines(false);

        JScrollPane jspSV = new JScrollPane();
        jspSV.setViewportView(serverFileView);

        jpHdlrStatus = new RFBHandlerJPanel();
        jpCardLOSrvView = new JPanel(new CardLayout());
        jpCardLOSrvView.add(jpHdlrStatus, FAKE_FT_VIEW);
        jpCardLOSrvView.add(jspSV, REAL_FT_VIEW);

        serverStatus = new JTextField("");
        serverStatus.setEditable(false);

        jpServerSide.add(cmdBtnsServer, BorderLayout.NORTH);
        jpServerSide.add(jpCardLOSrvView, BorderLayout.CENTER);
        jpServerSide.add(serverStatus, BorderLayout.SOUTH);

        showCard(false);
        // -------------------------
        // assemble the Status Panel
        // -------------------------
        JPanel statusPanel = new JPanel(new BorderLayout(blx, bly));
        jpProgBarEtc = new JPanel(new BorderLayout(blx, bly));
        stopButton = new JButton("Stop", new ImageIcon(this.getClass().getResource("/16x16/stop.png")));
        stopButton.addActionListener(this);
        stopButton.setAlignmentX(Component.CENTER_ALIGNMENT);
        stopButton.setVisible(true);
        stopButton.setEnabled(false);

        jProgressBar = new JProgressBar();
        jProgressBar.setPreferredSize(preferredDims4ProgyBar);
        jProgressBar.setStringPainted(true);

        jpProgBarEtc.add(stopButton, BorderLayout.WEST);
        jpProgBarEtc.add(jProgressBar, BorderLayout.CENTER);
        jpProgBarEtc.setVisible(false);


        statusPanel.add(jpProgBarEtc, BorderLayout.CENTER);

        setLayout(new BorderLayout(blx, bly));

        JPanel jpFileViews = new JPanel(new GridLayout(1, 2, blx, bly));
        jpFileViews.add(jpClientSide);
        jpFileViews.add(jpServerSide);

        JPanel jpFileViewsnButtons = new JPanel(new BorderLayout());
        jpFileViewsnButtons.add(jpFileViews, BorderLayout.CENTER);

        add(jpFileViewsnButtons, BorderLayout.CENTER);
        add(statusPanel, BorderLayout.SOUTH);

        updateClientDrivesComboBox();
        setPreferredSize(preferredDims);
    }


    private JTable getFTTable(boolean client) {
        return client ? clientFileView : serverFileView;
    }

    private FTTableDataModel getFTTableDM(boolean client) {
        return (FTTableDataModel) getFTTable(client).getModel();
    }

    private void setPreferredColumnSizes(boolean client) {
        TableColumn col = null;
        JTable t = getFTTable(client);
        TableColumnModel cm = t.getColumnModel();
        int cols = t.getColumnCount();

        for (int i = 0; i < cols; i++) {
            col = cm.getColumn(i);
            col.setPreferredWidth(i == 0 ? 200 : 120);
        }
    }

    private void onConnected() {
        showCard(true);
        serverFileView.setModel(new FTTableDataModel(hdlr.getRFS()));
        serverFileView.setDefaultRenderer(hdlr.getRFS().getColumnClass(0), new FSTableCellRenderer());
        setPreferredColumnSizes(false);

        TableRowSorter<FTTableDataModel> trsSrv = new TableRowSorter<FTTableDataModel>(getFTTableDM(false));
        List<RowSorter.SortKey> sortKeys = new ArrayList<RowSorter.SortKey>();
        sortKeys.add(new RowSorter.SortKey(0, SortOrder.ASCENDING));
        trsSrv.setSortKeys(sortKeys);
        trsSrv.setComparator(0, FTComparator.getFilenameComparator());
        trsSrv.setComparator(1, FTComparator.getSizeComparator());
        trsSrv.setComparator(2, FTComparator.getDateComparator());
        trsSrv.setComparator(3, FTComparator.getDateComparator());
        trsSrv.setComparator(4, FTComparator.getDateComparator());
        serverFileView.setRowSorter(trsSrv);
        trsSrv.sort();

        // This might be interesting someday when I need to filter the views contents
        // TableRowSorter<FTFileVector> trsSrv = new TableRowSorter<FTFileVector>( connection.getRFS());
        // serverFileView.setRowSorter( trsSrv);
        // trsSrv.setRowFilter( RowFilter.regexFilter( ".*A.*"));

        cmdBtnsServer.setLocation(hdlr.getHost());

        // Ensure the file-transfer preferences are being honored!
        if (ftPrefs != null) {
            cmdBtnsClient.bookmark.removeActionListener(this);
            cmdBtnsClient.bookmark.addBookmarks(ftPrefs.getClientBookmarks());
            cmdBtnsClient.bookmark.addActionListener(this);

            // addBookmarks much as visit_Path can fire selectionChanged
            // events which in turn can result in undesired changedir requests.
            cmdBtnsServer.bookmark.removeActionListener(this);
            cmdBtnsServer.bookmark.addBookmarks(ftPrefs.getServerBookmarks());
            cmdBtnsServer.bookmark.addActionListener(this);

            changeLocalDirectory(new File(ftPrefs.getPreferredClientDir()));

            onDTFormatChanged();

            // Ensure the users bookmarks can be persisted.
            mayExtractPreferences = true;
        }

        cmdBtnsServer.setFTBtnsEnabled(true);
        invalidateFTButtons();
        //VncViewer.pen.log( Level.CONFIG, "Connection has been established.");
    }

    private void onDisconnected() {
        showCard(false);
        cmdBtnsServer.setFTBtnsEnabled(false);
        cmdBtnsClient.btnTransfer.setEnabled(false);
        serverFileView.setEnabled(false);

        setFTBusy(false);
        jProgressBar.setValue(0);
        jpProgBarEtc.setVisible(false);

        //VncViewer.pen.log( Level.CONFIG, "Connection has been closed down.");
    }

    private void onFileDeletedEnded(RFBHandlerEvent e) {
        setFTBusy(false);
        reloadCurrentServerDir();
    }

    private void onDirectoryCreated(RFBHandlerEvent e) {
        setFTBusy(false);
        reloadCurrentServerDir();
    }

    private void onFileTransferBegin(RFBHandlerEvent e) {
        setFTBusy(true);
        if (e.isDownload) {
            abortErrorMsg = "User aborted download of [" + e.srcDownloadOrUnzip + "] to [" + e.dstDownloadOrUnzip + "]";
            jProgressBar.setIndeterminate(e.isDirectory);
            jProgressBar.setString("Waiting for server response");
            stopButton.setEnabled(true);
            jpProgBarEtc.setVisible(true);
        } else {
            abortErrorMsg = "User aborted upload of [" + e.srcDownloadOrUnzip + "] to [" + e.dstDownloadOrUnzip + "]";
            VNCViewer.pen.log(Level.FINEST, "onFileTransferBegin got called !!!!!!!!!!!");
            stopButton.setEnabled(true);
            jpProgBarEtc.setVisible(true);
        }
    }

    private void onFileTransferEnd(RFBHandlerEvent e) {
        VNCViewer.pen.log(Level.FINEST, "onFileTransferEnd got called !!!!!!!!!!!");

        abortErrorMsg = DEFABORTMSG;
        jProgressBar.setValue(0);
        stopButton.setEnabled(false);
        jpProgBarEtc.setVisible(false);

        if (e.success) {
            if (e.isDownload) reloadCurrentClientDir();
            else reloadCurrentServerDir();
        }
        setFTBusy(false);
    }

    private void onFileTransferError(RFBHandlerEvent e) {
        VNCViewer.pen.log(Level.FINEST, "onFileTransferError got called !!!!!!!!!!!");

        abortErrorMsg = DEFABORTMSG;
        stopButton.setEnabled(false);
        jProgressBar.setValue(0);
        jpProgBarEtc.setVisible(false);
        setFTBusy(false);
    }

    private void onFileTransferProgress(RFBHandlerEvent e) {
        jProgressBar.setIndeterminate(false);
        jProgressBar.setValue(e.progress);
        String fn = e.srcDownloadOrUnzip;
        String sDirection = e.isDownload ? "Download" : "Upload";
        String s = sDirection + " [" + fn + "] " + e.progress + "% complete @ [" + e.bps / 1024 + " kb/s]";
        jProgressBar.setString(s);
    }

    private void onZipProgress(RFBHandlerEvent e) {
        jProgressBar.setIndeterminate(false);
        jProgressBar.setValue(e.progress);
        String fn = e.srcDownloadOrUnzip;
        String sDirection = e.isDeflating ? "Zipping" : "Unzipping";
        String s = sDirection + " [" + fn + "] " + e.progress + "% complete";
        jProgressBar.setString(s);
    }

    private void onZipProcessStarted(RFBHandlerEvent e) {
        if (e.isDeflating) {
            jpProgBarEtc.setVisible(true);
        } else {
            jpProgBarEtc.setVisible(true);
        }
    }

    private void onZipProcessEnded(RFBHandlerEvent e) {
        if (e.isDeflating) {
            jpProgBarEtc.setVisible(false);
        } else {
            jpProgBarEtc.setVisible(false);
            reloadCurrentClientDir();
        }
    }

    private void onDirectoryChangeBegin(RFBHandlerEvent e) {
        setFTBusy(true);
        serverFileView.setEnabled(false);
        jProgressBar.setIndeterminate(true);
        jProgressBar.setString("Changing to directory [" + e.dstDownloadOrUnzip + "]");
        jpProgBarEtc.setVisible(true);
    }

    private void onDirectoryChangeEnd(RFBHandlerEvent e) {
        jpProgBarEtc.setVisible(false);
        updateRemoteFileSystemSummary();
        serverFileView.setEnabled(true);
        getFTTableDM(false).forceUpdate();
        setFTBusy(false);
        serverFileView.revalidate();
    }

    private void onRemoteFSGrew(RFBHandlerEvent e) {
        if (e == null || serverStatus == null) return;
        String fd = e.isDirectory ? "[directory]" : "[file]";
        serverStatus.setText("Added " + fd + " [" + e.srcDownloadOrUnzip + "]");
    }

    private void setFTBusy(boolean b) {
        if (b == hasBusyFTLine) return;
        hasBusyFTLine = b;
        invalidateFTButtons();
    }

    private void onDTFormatChanged() {
        if (ftPrefs == null) return;
        if (lfs != null) {
            getFTTableDM(true).setDateFormat(ftPrefs.getDateFormat());
            clientFileView.revalidate();
        }

        if (hdlr != null && hdlr.getRFS() != null) {
            getFTTableDM(false).setDateFormat(ftPrefs.getDateFormat());
            serverFileView.revalidate();
        }
    }

    /**
     * Implementation of the MouseListener Interface. Its mainly concerned with emitting
     * update requests for the remote and the local filesystem views.
     */
    public void mouseClicked(MouseEvent e) {
        // Mouse Double clicked
        if (e.getClickCount() == 2) {
            // Clicked on local file
            if (e.getSource() == clientFileView) changeToDirectorySelectedInClientFileView();

            // Clicked on remote file
            if (e.getSource() == serverFileView) changeToDirectorySelectedInServerFileView();
        }
    }

    public void mouseEntered(MouseEvent e) {
    }

    public void mouseExited(MouseEvent e) {
    }

    public void mousePressed(MouseEvent e) {
    }

    public void mouseReleased(MouseEvent e) {
    }

    /**
     * Implementation of the ListSelectionListener Interface. On selecting an
     * entry in either the servers or the clients directory list, the transfer
     * Button will be changed.
     */
    public void valueChanged(ListSelectionEvent e) {
        // We are not interested in interim events
        if (e.getValueIsAdjusting()) return;

        ListSelectionModel lsmServer = serverFileView.getSelectionModel();
        ListSelectionModel lsmClient = clientFileView.getSelectionModel();

        // Just to be on the safe side, we handle only events coming from our list views
        if (e.getSource() != lsmClient && e.getSource() != lsmServer) return;

        invalidateFTButtons(e.getSource() == lsmClient);
    }

    /**
     * Implementation of the ActionListener Interface. Handles all button
     * pressed Signals.
     */
    public void actionPerformed(ActionEvent evt) {
        if (evt.getSource() == cmdBtnsClient.btnTransfer)
            uploadSelectedFileOrDirectory();

        if (evt.getSource() == cmdBtnsServer.btnTransfer)
            downloadSelectedFileOrDirectory();

        if (evt.getSource() == cmdBtnsClient.btnReload)
            reloadCurrentClientDir();

        if (evt.getSource() == cmdBtnsServer.btnReload)
            reloadCurrentServerDir();

        if (evt.getSource() == cmdBtnsClient.btnRoot)
            gotoRoot(true);

        if (evt.getSource() == cmdBtnsServer.btnRoot)
            gotoRoot(false);

        if (evt.getSource() == cmdBtnsClient.btnDelete)
            doDelete(false);

        if (evt.getSource() == cmdBtnsServer.btnDelete)
            doDelete(true);

        if (evt.getSource() == cmdBtnsClient.btnNewFolder)
            doNewFolder(false);

        if (evt.getSource() == cmdBtnsServer.btnNewFolder)
            doNewFolder(true);

        if (evt.getSource() == cmdBtnsClient.bookmark)
            changeLocalDirectory(new File(cmdBtnsClient.bookmark.getCurrentLocation()));

        if (evt.getSource() == cmdBtnsServer.bookmark)
            hdlr.queue(RFBConnJob.changeDir(hdlr, cmdBtnsServer.bookmark.getCurrentLocation(), true));

        if (evt.getSource() == stopButton)
            doStop();
    }

    public void update(Observable o, Object arg) {
        if (arg instanceof RFBHandlerEvent) {
            if (o != hdlr) return;
            final RFBHandlerEvent e = (RFBHandlerEvent) arg;
            final Component c = this;

            try {
                // Since the update method is called by the RFB thread calls that
                // affect the GUI must be forcibly run on the GUI thread to avoid
                // undesired concurrency issues.
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        boolean isDC = false;

                        switch (e.operation) {
                            case NONE:
                                break;
                            case CONNECT:
                                onConnected();
                                break;
                            case DISCONNECT:
                                onDisconnected();
                                isDC = true;
                                break;
                            case DRVLISTCHANGED:
                                onDrvListChanged();
                                break;
                            case FTBEGIN:
                                onFileTransferBegin(e);
                                break;
                            case FTPROGRESS:
                                onFileTransferProgress(e);
                                break;
                            case FTEND:
                                onFileTransferEnd(e);
                                break;
                            case FTERROR:
                                onFileTransferError(e);
                                break;
                            case ZIPBEGIN:
                                onZipProcessStarted(e);
                                break;
                            case ZIPPROGRESS:
                                onZipProgress(e);
                                break;
                            case ZIPEND:
                                onZipProcessEnded(e);
                                break;
                            case CDBEGIN:
                                onDirectoryChangeBegin(e);
                                break;
                            case CDEND:
                                onDirectoryChangeEnd(e);
                                break;
                            case RFSGREW:
                                onRemoteFSGrew(e);
                                break;
                            case DIRCREATE:
                                onDirectoryCreated(e);
                                break;
                            case FILEDELETE:
                                onFileDeletedEnded(e);
                                break;
                            case PHASECHANGED:
                                jpHdlrStatus.repaint();
                                break;
                            // case FTBUSY         : onFTBusy( e); break;
                            // Intentionally ignoring these
                            case CLIPCHANGED:
                            case CONSOLESHUTDOWN:
                            case CURSORMOVED:
                            case FTBUSY:
                            case PDCHANGED:
                                break;
                            default:
                                break;
                        }

                        // Events indicating an operation failed are to be
                        // relayed to the user through a message box. Disconnects
                        // caused by server side timeouts however are not as they
                        // will likely confuse the user. They will be logged away.
                        if (!e.success && e.errMsg != null && !e.errMsg.isEmpty() && !isDC) {
                            JOptionPane.showMessageDialog(c, e.errMsg, FTERRR, JOptionPane.ERROR_MESSAGE);
                        }
                    }
                });
            } catch (Exception ex) {
            }
        }

        if (arg instanceof SettingsFT.FTProperty &&
                (arg == SettingsFT.FTProperty.DTFORMAT ||
                        arg == SettingsFT.FTProperty.DTFORMATSTRING)) {
            onDTFormatChanged();
        }
    }

    private void showCard(boolean real) {
        //System.err.println( "WILL BE SHOWING : " + (real ? REAL_FT_VIEW : FAKE_FT_VIEW));
        CardLayout l = (CardLayout) (jpCardLOSrvView.getLayout());
        l.show(jpCardLOSrvView, real ? REAL_FT_VIEW : FAKE_FT_VIEW);
    }

    private class FTTableDataModel implements TableModel {
        private Vector<TableModelListener> listeners = new Vector<TableModelListener>();
        private RemoteFileSystem rfs = null;
        private DateFormat df = new SimpleDateFormat("dd.MM.yyyy HH:mm");

        public FTTableDataModel(RemoteFileSystem rfs) {
            this.rfs = rfs;
        }

        public void setDateFormat(DateFormat f) {
            if (f != null)
                df = f;
            //setChanged( new TableModelEvent( this));
        }

        public void forceUpdate() {
            setChanged(new TableModelEvent(this));
        }

        @Override
        public void addTableModelListener(TableModelListener l) {
            listeners.add(l);
        }

        @Override
        public Class<?> getColumnClass(int columnIndex) {
            switch (columnIndex) {
                case 0:
                    return FTFile.class;
                case 1:
                    return FTLongColumn.class;
                case 2:
                    return FTDateColumn.class;
                case 3:
                    return FTDateColumn.class;
                case 4:
                    return FTDateColumn.class;
                default:
                    return String.class;
            }
        }

        @Override
        public int getColumnCount() {
            return 5;
        }

        @Override
        public String getColumnName(int columnIndex) {
            switch (columnIndex) {
                case 0:
                    return ("Filename");
                case 1:
                    return ("Size (B)");
                case 2:
                    return ("Write");
                case 3:
                    return ("Access");
                case 4:
                    return ("Created");

                default:
                    return ("Error");
            }
        }

        @Override
        public int getRowCount() {
            return rfs.size();
        }

        @Override
        public Object getValueAt(int rowIndex, int columnIndex) {
            FTFile f;

            try {
                f = rfs.elementAt(rowIndex);
            } catch (ArrayIndexOutOfBoundsException e) {
                f = FTFile.PLACEHOLDER;
            }

            switch (columnIndex) {
                case 0:
                    return f;
                case 1:
                    return new FTLongColumn(f.getFilename(), f.getSize());
                case 2:
                    return new FTDateColumn(f.getFilename(), f.getLastWrittenToOn(), df);
                case 3:
                    return new FTDateColumn(f.getFilename(), f.getLastAccessedOn(), df);
                case 4:
                    return new FTDateColumn(f.getFilename(), f.getCreatedOn(), df);

                default:
                    return ("Error");
            }
        }

        @Override
        public boolean isCellEditable(int arg0, int arg1) {
            return false;
        }

        @Override
        public void removeTableModelListener(TableModelListener l) {
            listeners.remove(l);
        }

        @Override
        public void setValueAt(Object arg0, int arg1, int arg2) {
        }

        protected void setChanged(TableModelEvent e) {
            Iterator<TableModelListener> itTML = listeners.iterator();

            while (itTML.hasNext()) {
                itTML.next().tableChanged(e);
            }
        }

    }
}
