package com.wincornixdorf.vnc.rfb;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.util.logging.Level;

import com.wincornixdorf.vnc.global.VNCViewer;

public class FBSFileSocket extends Socket {

    private class DevNullOutputStream extends OutputStream {
        public DevNullOutputStream() { super(); }
        public void close() {};
        public void flush() {};
        public void write( byte[] b) {};
        public void write( byte[] b, int off, int len) {};
        public void write( int b) {};
    }

    // Doesn't work as desired.
    private class LoopingFileInputStream extends BufferedInputStream {
        public LoopingFileInputStream( String fn) throws IOException {
            super( new FileInputStream( fn));
        }

        public int read() throws IOException {
            int i = -1;
            try {
                i = super.read();
/*                if ( i == -1) {
                    super.reset();
                    i = super.read();
                }*/
            } catch ( IOException e) {
                e.printStackTrace();
                super.reset();
                i = super.read();
            } 
            return i;
        }

        public int read( byte[] b) throws IOException {
            int i = -1;
            try {
                i = super.read( b);
                if ( i == -1) {
                    super.reset();
                    i = super.read( b);
                }
            } catch ( IOException e) {
                e.printStackTrace();
                super.reset();
            } 
            return i;
        }

        public int read( byte[] b, int off, int len) throws IOException {
            int i = -1;
            try {
                i = super.read( b, off, len);
                if ( i == -1) {
                    super.reset();
                    i = super.read( b, off, len);
                }
            } catch ( IOException e) {
                e.printStackTrace();
                super.reset();
            }
            return i;
        }

        public long skip( long n) throws IOException {
            long l = 0;
            try {
                l = super.skip( n);
            } catch ( IOException e) {
                super.reset();
            }
            return l;
        }
    }

    private OutputStream    os          = null;
    private InputStream     is          = null;

    public FBSFileSocket( String fn) throws IOException {
        VNCViewer.pen.log( Level.FINEST, "FBSFileSocket(" + fn + ")!");
        os = new DevNullOutputStream();
        is = new LoopingFileInputStream( fn);
    }

    public void close() throws IOException {
        os.close();
        is.close();
    }

    public OutputStream getOutputStream() throws IOException {
        return os;
    }

    public InputStream getInputStream() throws IOException {
        return is;
    }

    public void setLoopMark() {
        is.mark( Integer.MAX_VALUE);
    }

    public boolean isBound() { return true; }

    public boolean isClosed() { return true; }
    public boolean isConnected() { return true; }
}
