/**
 * 
 */
package com.wincornixdorf.vnc.rfb;

import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.util.Iterator;
import java.util.Vector;
import java.util.concurrent.Callable;
import java.util.logging.Level;

import javax.net.ssl.SSLParameters;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSocketFactory;

import com.wincornixdorf.utils.httpSocket.HTTPClientSocket;
import com.wincornixdorf.vnc.global.VNCViewer;
import com.wincornixdorf.vnc.rfb.conf.VNCSrvCfg;
import com.wincornixdorf.vnc.rfb.conf.VNCSrvCfg.ConMode;

/**
 * @author dic
 *
 */
public class SocketForge implements Callable<Socket> {

    // If we don't have a connection with the server within 60 seconds, it
    // is safe to assume the server is down or someone switched to TCP/IP
    // by avian carrier. ( It works! Ask that pigeon on your window sill.)
    private static final int      socketTimeout   = 120000;

    private VNCSrvCfg       cfg             = null;
    private String          connType        = "error";
    
    public SocketForge( VNCSrvCfg cfg, boolean forgeFBConn) {
        this.cfg            = cfg;
        this.connType       = forgeFBConn ? "FB" : "FT";
    }

    private String logStr( String s) { return( "[" + connType + "] " + s); }

    private Socket forge( VNCSrvCfg.ConMode conMode, InetSocketAddress isa) throws IOException {        
        Socket                  socketForged    = null;
        
        InetAddress addr = isa != null ? isa.getAddress() : null; 
        if ( addr == null) {
        	VNCViewer.pen.log( Level.WARNING, logStr( "Failed to forge a Socket: No address given!"));
        	return socketForged;
        }
        
        String hs = "Trying to \"connect\" to ";
        
        switch ( conMode) {
            case FILE :
                VNCViewer.pen.log( Level.WARNING, logStr( hs + "File [" + cfg.filename + "]!"));
                socketForged = new FBSFileSocket( cfg.filename);
                // // // giveFlag( REPLAYING);
            break;
            case PLAIN :
                VNCViewer.pen.log( Level.WARNING, logStr( hs + "Plain Socket [" + isa + "]!"));
                socketForged = new Socket( addr, isa.getPort());
                socketForged.setSoTimeout( socketTimeout);
            break;
            case SSL :
            case TLS :
                VNCViewer.pen.log( Level.CONFIG, logStr( hs + "TLS Socket [" + isa + "]!"));

                SSLSocketFactory factory = VNCViewer.sslContext.getSocketFactory();
                SSLSocket sslSock = (SSLSocket)factory.createSocket();

                SSLParameters sslpar = new SSLParameters();

                // Restricting the socket to these protocols. This is to make sure we won't crater if
                // SSLv2Hello hasn't been enabled on the Servers side. At least that's the idea :/
                //String[] protocols = { "SSLv2Hello", "TLSv1", "SSLv3" };
                
                // In light of the recent "Poodle" attack https://www.openssl.org/~bodo/ssl-poodle.pdf
                // SSLv3 support had to be canceled as well. Good riddance. That protocol is almost 15
                // years old now anyways.
                
                String[] protocols = getProtocols();
                sslpar.setProtocols( protocols);
                sslSock.setSSLParameters( sslpar);

                // Warning! This line might pose a problem someday....
                sslSock.setEnabledCipherSuites( sslSock.getSupportedCipherSuites());
                sslSock.setSoTimeout( socketTimeout);

                sslSock.connect( isa, socketTimeout);
                sslSock.startHandshake();
                socketForged = sslSock;

                //SSLSession mySSLsess = sslSock.getSession();

                // Retrieve the server's certificate chain
                //java.security.cert.Certificate[] serverCerts =
                //socket.getSession().getPeerCertificates();

                //VncViewer.pen.log( Level.CONFIG, logStr( "--------Using a SSL secured Connection--------"));
                //VncViewer.pen.log( Level.CONFIG, logStr( "Cipher  : " + mySSLsess.getCipherSuite()));
                //VncViewer.pen.log( Level.CONFIG, logStr( "PeerHost: " + mySSLsess.getPeerHost()));
                //VncViewer.pen.log( Level.CONFIG, logStr( "Protocol: " + mySSLsess.getProtocol()));
                //java.security.cert.Certificate[] lcerts = mySSLsess.getLocalCertificates();
                //for ( int i = 0; i < ((lcerts != null) ? lcerts.length : 0); i++) {
                //    VncViewer.pen.log( Level.CONFIG, logStr( "Cert sent: " + lcerts[ i]));
                //}
                //VncViewer.pen.log( Level.CONFIG, logStr( "----------------------------------------------"));
            break;
            case HTTP :
                VNCViewer.pen.log( Level.CONFIG, logStr( hs + "HTTP Socket [" + isa + "]!"));
                socketForged = new HTTPClientSocket( addr.getHostAddress(), isa.getPort(), false, connType, null);
            break;
            case HTTPS :
                VNCViewer.pen.log( Level.WARNING, logStr( hs + "HTTPS Socket [" + isa + "]! THIS SOCKET DOES NOT HAVE A SSLContext yet!!!"));
                socketForged = new HTTPClientSocket( addr.getHostAddress(), isa.getPort(), true, connType, null);
            break;
            default :
        }
        
        return socketForged;
    }

    private String[] getProtocols() {
		return new String[] {"TLSv1"};
	}

	private Socket tryForgeASocket() throws IOException {
    	Socket            s       = null;
    	VNCSrvCfg.ConMode conMode = cfg.getConMode();

    	if ( conMode == ConMode.FILE) return forge( ConMode.FILE, null);

    	Vector<InetSocketAddress> v = cfg.get1stHops();
    	Iterator<InetSocketAddress> it = v.iterator();
    	
    	while ( s == null && it.hasNext()) {
        	InetSocketAddress isa = it.next();
    		s = forge( conMode, isa);
    		if ( s == null || !s.isConnected()) {
    			VNCViewer.pen.log( Level.WARNING, logStr( "[" + conMode + "] Socket [" + isa.toString() + "] was unusable. Looking for alternatives..."));
    			s = null;
    		}
    	}

		return s;
	}
    
    @Override
    public Socket call() throws Exception {
        if ( cfg == null) throw new IllegalArgumentException( "SocketForge has not been initialized!");
        return tryForgeASocket();
    }


}
