package com.wincornixdorf.vnc.rfb;

import com.wincornixdorf.vnc.global.VNCViewer;
import com.wincornixdorf.vnc.rfb.conf.VNCSessionCfg;
import com.wincornixdorf.vnc.rfb.conf.VNCSrvCfg;
import com.wincornixdorf.vnc.rfb.debug.RFBHandlerObserver;

import java.io.File;
import java.util.Observable;
import java.util.Observer;
import java.util.logging.Level;

public class VNCSession extends Observable implements Observer {

//    /**A reference to the VNCSession pool this Session is contained in.*/
//    private VNCSessionPool      pool                    = null;

    private VNCSessionCfg cfg = null;

    /**
     * This connection is supposed to be used in all things related to the
     * remote framebuffer. It will not accept any file transfer related calls.
     */
    private RFBHandler connection = null;

    /**
     * The thread handling the Framebuffer related network traffic.
     */
    private Thread runThread = null;

    /**
     * The parameters used to configure the connection. Among others the
     * hostname/username etc.
     */
    private VNCSrvCfg srvCfg = null;

    private String sid = null;

    //    public VNCSession( VNCSessionPool pool, VNCSessionCfg cfgIn, String sidIn)
    public VNCSession(VNCSessionCfg cfgIn, String sidIn) {
//        this.pool   = pool;
        this.srvCfg = cfgIn.vncserver;
        this.cfg = cfgIn;
        this.sid = sidIn;

        createConn();
    }

    public String getSID() {
        return sid;
    }

    public VNCSessionCfg getConfiguration() {
        return cfg;
    }

    public RFBHandler getFB() {
        return connection;
    }

    public RFBHandler getFT() {
        return connection;
    }

    public VNCSrvCfg getHost() {
        return srvCfg;
    }

    /**
     * @return True if the session has no open connections.
     */
    public boolean hasBothConnectionsDown() {
        return (!hasOpenConnection(true) && !hasOpenConnection(false));
    }

    /**
     * @return True if the session has got at least one open connection.
     */
    public boolean hasOpenConnection() {
        return (hasOpenConnection(true) || hasOpenConnection(false));
    }

    /**
     * @return True if the sessions connections are not in the state they
     * are supposed to be.
     */
    public boolean hasBeenCrippled() {
        // For now in a session both connections must be up
        return !hasOpenConnection();
    }

    /**
     * Used to establish whether a connection is up or not.
     *
     * @param fb Set true, if interested int the framebuffer connection, false if it's the file transfer connection.
     * @return True if the specified connection is open.
     */
    public boolean hasOpenConnection(boolean fb) {
/*        if ( fb) return( connection != null ? connection.hasFlag( VNCConnection.RUNNING) : false);
        else return( connFT != null ? connFT.hasFlag( VNCConnection.RUNNING) : false);*/
        return (connection != null ? !connection.isDeadInTheWater() : false);

    }

    /**
     * @return True if the sessions FB connection is open.
     */
    public boolean supportsRemoteControl() {
        return (hasOpenConnection(true));
    }

    private void createConn() {
        if (connection != null) {
            VNCViewer.pen.log(Level.WARNING, "Can't create more than one connection.");
            return;
        }

        connection = new RFBHandler(this, cfg.cn.getVNCMode());
        connection.addObserver(this);

        setUpDebugTrace(connection, "FBHandler");
        runThread = new Thread(connection);
        runThread.setName("FBHdlr_Thread");
        runThread.setPriority(Thread.NORM_PRIORITY);
        runThread.start();

//        ExecutorService executor = Executors.newFixedThreadPool(1);
//        FutureTask<Socket> sockSmith = new FutureTask<Socket>( new SocketForge( cfg.vncserver, true));

//        executor.execute( sockSmith);
//        try
//        {
//            System.out.println("Here is the result..." + sockSmith.get());
//        } catch (ExecutionException ex) {}

//        runThread = new Thread( new Runnable() {
//            public void run() {
//                SocketForge sf = new SocketForge( cfg.vncserver, true);
//                try {
//                    Socket s = sf.call();
//                } catch ( Exception e) {
//                    e.printStackTrace();
//                }
//                //set up socket
//                // new VNCConnection( this, socket.getInput, socket.getOutput());
//            }
//        });

    }

    public String toString() {
        if (cfg == null) return "Error";
        return cfg.toString();
    }

    public void update(Observable o, Object arg) {
        // The VNCSession is "forwarding" all VNCConnectionEvents coming
        // from its connection objects, provided they signal gaining or
        // loosing the connected status or changes to the clipboard contents.
        RFBHandlerEvent e = (RFBHandlerEvent) arg;
        if (e.operation == RFBHandlerEvent.Category.CONNECT ||
                e.operation == RFBHandlerEvent.Category.DISCONNECT ||
                e.operation == RFBHandlerEvent.Category.CLIPCHANGED) {
            setChanged();
            notifyObservers(e);
        }

/*        if ( o == connection &&
             e.operation == VNCConnectionEvent.Category.STATUSCHANGED &&
             connection.hasFlag( VNCConnection.RUNNING))
        {
            //TODO: Write code along the lines: if FB came up and FT should be up too but isn't, then open the connection.
        }*/
    }

    /**
     * Set up a debug trace file that will persist RFBHandlers messages.
     */
    private void setUpDebugTrace(RFBHandler h, String name) {
        if (!VNCViewer.isDivine()) return;

        String propTrace = System.getProperty("TRACEVNC");

        if (propTrace == null || propTrace.isEmpty()) return;

        File f = new File(System.getProperty("user.home") +
                File.separatorChar + "tmp" +
                File.separatorChar + name);
        RFBHandlerObserver o = new RFBHandlerObserver(f, name);
        h.addObserver(o);
    }
}
