package com.wincornixdorf.vnc.rfb;


import java.util.logging.Level;

import javax.swing.DefaultListModel;

import com.wincornixdorf.vnc.global.VNCViewer;
import com.wincornixdorf.vnc.rfb.conf.VNCSessionCfg;


public class VNCSessionPool extends DefaultListModel<VNCSession>
{
    private static final long   serialVersionUID    = 3577038029659166374L;
    private static final int    maxDivine           = 10;
    private static final int    maxNormal           = 10;
    
    private int                 counter             = 0;

    public VNCSessionPool() {
        super();
    }

    public boolean isFull() {
        return( getSize() >= getMaxSessions());
    }

    public synchronized VNCSession spawnSession( VNCSessionCfg cfg) {
        if ( isFull()) {
            VNCViewer.pen.log( Level.WARNING, "Failed to open a new Session. The session pool is full.");
            return null;
        }

        if ( cfg == null || cfg.vncserver == null || !cfg.vncserver.isConfigured()) {
            VNCViewer.pen.log( Level.WARNING, "Failed to open a new Session. The configuration is incomplete.");
            return null;
        }

        VNCViewer.pen.log( Level.FINEST, "\n" + cfg.vncserver.toUserFriendlyString());

        String sid = "Session [" + getCounter( true) + "]";
        VNCSession sess = new VNCSession( cfg, sid);
        addElement( sess);
        return sess;
    }

    public synchronized void destroySession( VNCSession s) {
        if ( s == null) {
            VNCViewer.pen.log( Level.WARNING, "Could not close session! None specified!");
            return;
        }

        VNCViewer.pen.log( Level.WARNING, "Closing Session: [" + s.toString() + "]");

        RFBHandler cFB = s.getFB();
        RFBHandler cFT = s.getFT();

        if ( cFB != null) {
        	VNCViewer.pen.log( Level.WARNING, "Sending kill thread request to FB connection!");
        	cFB.queue( RFBConnJob.terminateThread( cFB));
        } else VNCViewer.pen.log( Level.WARNING, "FAILED to send kill thread request to FB connection!");

        if ( cFT != null) {
        	VNCViewer.pen.log( Level.WARNING, "sending kill thread request to FT connection!");
        	cFT.queue( RFBConnJob.terminateThread( cFT));
        } else VNCViewer.pen.log( Level.WARNING, "FAILED to send kill thread request to FT connection!");


        // I am sleeping a little to give the connections time to dismantle.
//        try { Thread.sleep( 300); } catch ( Exception ex) {}

        //         TODO: Gotta look into this. Probably removeElement will destroy the item on the heap without giving the connection time to close down!
        removeElement( s);
    }
    
    private synchronized int getCounter( boolean next) {
        if ( !next) return counter;
        else return counter++;
    }

    private int getMaxSessions() {
        return VNCViewer.isDivine() ? maxDivine : maxNormal;
    }
}
