package com.wincornixdorf.vnc.rfb;

import java.util.EventObject;

public class VNCSessionPoolEvent extends EventObject
{
    private static final long serialVersionUID = -2945367665779970874L;

    public enum Category {
        NONE,           /**Merely a fallback.*/
        ADDED,          /**A new Session was created. Connections are down though.*/
        REMOVED,        /**A new Session was destroyed.*/
        ACTIVATED       /**The Session is now the Main session*/
    }

    /**The kind of operation this Event is referring to.*/
    public Category operation   = Category.NONE;

    /**Will be false until the operation was successfully completed.*/
    public boolean success          = false;

    /**In the event of a REMOVED event this flag will be true, if the last
    Session object was removed.*/
    public boolean dry              = false;

    /**The Session which is at the center of the event.*/
    public VNCSession session       = null;

    /**The Session which is not at the center of the event but significantly involved.
    E.g. when activating a session, the old session will be referenced here.*/
    public VNCSession oldSession    = null;

    public static String toString( Category c) {
        switch( c) {
            case NONE           : return( "NONE");
            case ADDED          : return( "ADDED A NEW SESSION");
            case REMOVED        : return( "REMOVED AN EXISTING SESSION");
            case ACTIVATED      : return( "SESSION MADE MAIN SESSION");
            default             : return( "ERROR");
        }
    }

    public VNCSessionPoolEvent( VNCSessionPool source) {
        super( source);
    }

    public void dump() {
        System.out.println( "----------- VNCSessionPoolEvent ------------");
        System.out.println( "============================================");
        //System.out.println( "Connected to: [" + session.getConnectionDetails() + "]");
        System.out.println( "success     : [" + success                        + "]");
    }
}
