package com.wincornixdorf.vnc.rfb.auth;

import java.io.DataInputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.spec.InvalidKeySpecException;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;
import java.util.logging.Level;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.DESKeySpec;
import javax.crypto.spec.IvParameterSpec;

import com.wincornixdorf.vnc.global.VNCViewer;
import com.wincornixdorf.vnc.rfb.RFB;
import com.wincornixdorf.vnc.rfb.RFB.AuthType;
import com.wincornixdorf.vnc.rfb.RFBHandler;
import com.wincornixdorf.vnc.rfb.RFB.AuthResult;
import com.wincornixdorf.vnc.rfb.conf.VNCSrvCfg;
import com.wincornixdorf.vnc.rfb.debug.ToolBox;
import com.wincornixdorf.vnc.rfb.exceptions.LoginException;

/**This class is responsible for Login and Authentication (LA) to a VNC Server.
It serves solely towards clarifying the code, keeping it manageable.*/
public class RFBHandlerLA {
    // These are SO PRIVATE! Hands off or my wrath shall be eternal!
    /**This is a COPY of the sacred stream the session uses to read from the socket.
    Come heaven, hell high waters or a newsflash saying hell had just frozen over
    DO NOT SHARE THIS STREAM WITH ANY OTHER THREAD!*/
    private DataInputStream is                 = null;
    /**This is a COPY of the sacred stream the session uses to write to the socket.
    Come heaven, hell high waters or a newsflash saying hell had just frozen over
    DO NOT SHARE THIS STREAM WITH ANY OTHER THREAD!*/
    private OutputStream    os                 = null;

    /**Used 4 internal reasons only. Will probably be removed soon. Do NOT use!*/
    private RFBHandler   vc                 = null;
    
    public RFBHandlerLA( RFBHandler s) {
        this.vc = s;
    }

    public void setStreams( DataInputStream isOfSession, OutputStream osOfSession) {
        this.is = isOfSession;
        this.os = osOfSession;
    }

    public boolean tryToConnect()
        throws LoginException
    { 
        boolean loggedIn = false;
        readVersionMsg();
        VNCViewer.pen.log( Level.FINEST, vc.logStr( "RFB server supports protocol version [" + vc.getServerParams().getVersionMajor( true) + "." + vc.getServerParams().getVersionMinor( true)) + "]");

        writeVersionMsg();
        VNCViewer.pen.log( Level.FINEST, vc.logStr( "Using RFB protocol version [" + vc.getServerParams().getVersionMajor( false) + "." + vc.getServerParams().getVersionMinor( false)) + "]");

        // authMode is former secType
        RFB.AuthType authType = negotiateModeOfAuthentification();
        VNCViewer.pen.log( Level.FINEST, vc.logStr( "Using secType [" + authType + "] " + authType.toDbgStr()));

		//        // If the server supports the Tight authentication, it's safe to
		//        // presume the server will understand the tight specific enhancements too.
		//        if ( secType == RFB.SecTypeTight) {
		//            // VncViewer.pen.log( Level.WARNING, vc.logStr( "Enabling TightVNC protocol extensions"));
		//            // For the time being, Tight extensions will not be used.
		//            // rfb.initCapabilities();
		//            // rfb.setupTunneling();
		//            //authType = negotiateAuthenticationTight();
		//            VncViewer.pen.log( Level.FINEST, vc.logStr( "Client and Server agreed on using auth type: [" + authTypeRequested + "]"));
		//        } else {
		//            authTypeRequested = secType;
		//            VncViewer.pen.log( Level.FINEST, vc.logStr( "The server demands auth type: [" + authTypeRequested + "]"));
		//        }
		
		            //mslogon = vc.getServerParams().getVersionMinor( true) == 4;
		/*            if ( vc.getServerParams().getVersionMinor( true) == 4) {
		                mslogon = true;
		                vc.logWarn( "Ultr@VNC mslogon detected");
		            }*/
		
		//             authType = RFB.AuthenticationMSLogon;

        switch ( authType) {
            case NONE:
                // No authentication needed
                VNCViewer.pen.log( Level.WARNING, vc.logStr( "Server does not require authentification!"));
                loggedIn = authenticateNone();
            break;
            case VNCPASSWORD:
                VNCViewer.pen.log( Level.CONFIG, vc.logStr( "VNC Password authentification detected."));
                loggedIn = authenticateVNC( vc.getServerParams().getHostParams().getPassword());
            break;
            case MSLOGONII:
            case MSLOGON:
                VNCViewer.pen.log( Level.CONFIG, vc.logStr( "MS-Logon (DH) detected."));
                loggedIn = authenticateMSDomain( vc.getServerParams().getHostParams().getUsername(), vc.getServerParams().getHostParams().getPassword());
            break;
            case ULTRA:
                VNCViewer.pen.log( Level.CONFIG, vc.logStr( "New Ultr@VNC DSM plugin based authetification detected."));
            break;
            default:
                 throw new LoginException( "Unknown authentication mode [" + authType + "]");
        }
        return loggedIn;
    }

    private RFB.AuthType negotiateModeOfAuthentification() throws LoginException {
        return vc.getServerParams().getVersionMinor( false) >= 7 ?
               pickBestAuthModeOfferedByTheServer() :
               readAuthModeRequiredByServer();
    }
    
    private RFB.AuthType readAuthModeRequiredByServer() throws LoginException {
        VNCViewer.pen.log( Level.FINEST, vc.logStr( "Reading security type"));
        
        int secType = 0;
		try {
			secType = is.readInt();
		} catch ( IOException e) {
			throw new LoginException( "Server failed to send the security type it wants to use.", e);
		}
        VNCViewer.pen.log( Level.FINEST, vc.logStr( "Read security type [" + secType + "]"));
        return RFB.AuthType.byValue( (byte)secType);
    }

    private List<RFB.AuthType> readAuthTypesOfferedByServer( boolean includeUnsupported)
    		throws LoginException
    {
        Vector<RFB.AuthType> v = new Vector<RFB.AuthType>();

        // The server will 1st send the number of authentication types it
        // understands and knows as an unsigned byte. If the server replies
        // with 0, something went wrong.
        int nAuthTypes = 0;
        
		try {
			nAuthTypes = is.readUnsignedByte();
		} catch ( IOException e) {
			throw new LoginException( "Server would not offer any auth types.", e);
		}
		
        if ( nAuthTypes <= 0) {
			readConnFailedReason( "Server didn't offer any authentication modes. (This may hint at a Repeater failure)", AuthResult.FAILED);
            return v;
        } else
            VNCViewer.pen.log( Level.FINEST, vc.logStr( "The Server offered [" + nAuthTypes + "] security types!"));

        // Otherwise the authentication types are being sent represented as
        // unsigned bytes, which then have to be interpreted as per the protocol.
        RFB.AuthType at = RFB.AuthType.INVALID;
        byte[] authTypes = new byte[ nAuthTypes];
        
        try {
        	if(vc.getServerParams().isRemoteSupport()){
        		// buggy SCIII sends wrong data - we have to read it and ignore it
        		is.read();
        		return v;
        	} else{
        		is.readFully( authTypes, 0, authTypes.length);
        	}
		} catch ( IOException e) {
			throw new LoginException( "", e);
		}

        String hs = "";

        for ( int i = 0; i < nAuthTypes; i++) {
            at = RFB.AuthType.byValue( authTypes[ i]);
            hs = "(" + ( i + 1) + "/" + nAuthTypes + ") [" + at.toDbgStr() + "] ";
            if ( includeUnsupported || at.isSupported()) {
                VNCViewer.pen.log( Level.FINEST, vc.logStr( hs + "Is supported."));
                v.add( at);
            } else VNCViewer.pen.log( Level.FINEST, vc.logStr( hs + "Is NOT supported. => Ignored"));
        }

        listSecTypes( v, "Server is offering [" + nAuthTypes + "] mode(s) of authentification:");

        return v;
    }
    
    /**Select security type from the server's list (protocol versions 3.7/3.8). 
     * @throws ConnFailedException */
    private RFB.AuthType pickBestAuthModeOfferedByTheServer() throws LoginException {
        RFB.AuthType curr = RFB.AuthType.INVALID;
        RFB.AuthType best = RFB.AuthType.INVALID;

        VNCViewer.pen.log( Level.FINEST, vc.logStr( "Reading list of security types supported by server."));
        List<RFB.AuthType> atAvailable = readAuthTypesOfferedByServer( false); // false = skip modes !supported by this viewer
        if(!vc.getServerParams().isRemoteSupport()){
        // Pick the security type we like best
        VNCViewer.pen.log( Level.FINEST, vc.logStr( "Picking the security type we like best."));
        Iterator<RFB.AuthType> it = atAvailable.iterator();
        
        while ( it.hasNext()) {
            curr = it.next();
            
            if ( curr.isPreferable( best)) best = curr;
        }
        } else {
        best = AuthType.ULTRA;
        }
        VNCViewer.pen.log( Level.FINEST, vc.logStr( "Will use [" + best.toDbgStr() + "]."));

        if ( best != RFB.AuthType.INVALID) {
			try {
		        VNCViewer.pen.log( Level.FINEST, vc.logStr( "Telling Server to use [" + best.toDbgStr() + "]."));
				os.write( best.asInt());
				os.flush();
				VNCViewer.pen.log( Level.FINEST, vc.logStr( "Told the Server to use [" + best.toDbgStr() + "]."));
			} catch ( IOException e) {
				throw new LoginException( "Failed to send the server the desired auth type (" + best.asInt() + "/" + best.toString() + ").", e);
			}
        } else
        	throw new LoginException( "None of the security types offered by the Server where supported!");

        return best;
    }

    private void listSecTypes( List<RFB.AuthType> secTypes, String caption) {
        // The security types will be logged only if using the most verbose log level!
        if ( secTypes == null || !VNCViewer.pen.isLoggable( Level.FINEST)) return;
        
        VNCViewer.pen.log( Level.FINEST, vc.logStr( caption));
        
        int no = 1;
        RFB.AuthType           at = RFB.AuthType.INVALID;
        Iterator<RFB.AuthType> it = secTypes.iterator();
        
        while ( it.hasNext()) {
            at = it.next();
            VNCViewer.pen.log( Level.FINEST, vc.logStr( "[#" + (no++) + "] " + at + " = " + at.toDbgStr()));
        }
    }

    private boolean authenticateNone() throws LoginException {
        if ( vc.getServerParams().getVersionMinor( false) >= 8)
            return isLoginSuccessful( "Passwordless");
        return true;
    }

    private boolean authenticateVNC( String pw) throws LoginException {
        byte[] challenge = new byte[ 16];
        
        try {
			is.readFully( challenge);
		} catch ( IOException e) {
			throw new LoginException( "VNC password login failed: Couldn't read the servers VNC authentication challenge.", e);
		}
        
        if ( pw.length() > 8)
            pw = pw.substring( 0, 8); // Truncate to 8 chars

        // Truncate password on the first zero byte.
        int firstZero = pw.indexOf( 0);
        if ( firstZero != -1)
            pw = pw.substring( 0, firstZero);

        byte[] key = {0, 0, 0, 0, 0, 0, 0, 0};
        System.arraycopy( pw.getBytes(), 0, key, 0, pw.length());

        try {
        	byte[] h = new byte[ 8];
        	System.arraycopy( challenge, 0, h, 0, 8);
			os.write( encryptVNCDES( h, key, true, false));
        	System.arraycopy( challenge, 8, h, 0, 8);
			os.write( encryptVNCDES( h, key, true, false));
		} catch ( Exception e) {
			throw new LoginException( "Failed to compose response to the servers challenge!", e);
		}

        return isLoginSuccessful( "VNC authentication");
    }

    public static byte[] longToBytes(long number) {
        byte[] bytes  = new byte[8];
        for (int i = 0; i < 8; i++) {
            bytes[i] = (byte) (0xff & (number >> (8 * (7 - i))));
        }
        return bytes;
    }

    private static byte flipBitsInByte( byte b) {
        return (byte) ( Integer.reverse( b) >>> (Integer.SIZE - Byte.SIZE));
    }

    public static byte[] flipBitsBytewise( byte[] bArr) {
    	byte rval[] = new byte[ bArr.length]; 

    	for ( int i = 0; i < bArr.length; i++) {
    		rval[ i] = flipBitsInByte( bArr[ i]);
    	}

    	return rval;
    }

    public static byte[] getAlignedBytes( byte[] zeInput) {
        int lenInput = zeInput.length;
        int lenAlign = ((int)((lenInput - 1) / 8) + 1) * 8;
        
        byte[] rOutput = new byte[ lenAlign];
        System.arraycopy( zeInput, 0, rOutput, 0, lenInput);
        for ( int i = lenInput; i < lenAlign; i++) rOutput[ i] = 0;

        return rOutput;
    }

    /**This method will subject the input to a DES with cipher block chaining but no padding.
     * The bits of every key byte will first be reversed, before they are used with the cipher.
     * This is a peculiarity of VNC. Which has been discussed at http://www.vidarholen.net/contents/junk/vnc.html.
     * @param input The bytes to be encrypted.
     * @param key   The key, which has yet to be bit-flipped to be used.
     * @return The DES encrypted input using the key with the bits of each byte reversed for input.*/
    private static byte[] encryptVNCDES( byte[] input, byte[] key, boolean flip, boolean useCBC)
    		throws InvalidKeyException,
    		       NoSuchAlgorithmException,
    		       InvalidKeySpecException,
    		       NoSuchPaddingException,
    		       InvalidAlgorithmParameterException,
    		       IllegalBlockSizeException,
    		       BadPaddingException
    {
        byte[] ivBytes = new byte[ 8];
        byte[] keyBytes = flip ? flipBitsBytewise( key) : key;
        byte[] in 		= getAlignedBytes( input);

        System.arraycopy( key, 0, ivBytes, 0, 8);
        
        DESKeySpec desKeySpec = new DESKeySpec( keyBytes);
        SecretKeyFactory keyFactory = SecretKeyFactory.getInstance( "DES");
        SecretKey secretKey = keyFactory.generateSecret( desKeySpec);

        String mode = useCBC ? "CBC" : "ECB";
        Cipher cipher = Cipher.getInstance( "DES/" + mode + "/NoPadding");
        if ( useCBC) cipher.init( Cipher.ENCRYPT_MODE, secretKey, new IvParameterSpec( ivBytes));
        else cipher.init( Cipher.ENCRYPT_MODE, secretKey);

        return cipher.doFinal( in);
    }

    private boolean authenticateMSDomain( String us, String pw) throws LoginException
    {
        if ( us == null || pw == null)
            throw new LoginException( "MS-Logon requires a username and a password.");
        
        byte[] zeGen 	= new byte[ 8]; 
        byte[] zeMod 	= new byte[ 8]; 
        byte[] zeKey 	= new byte[ 8]; 
        String excMsg 	= "generator";

        try {
            is.readFully( zeGen);
            excMsg = "modulus";
            is.readFully( zeMod);
            excMsg = "itermediate key";
            is.readFully( zeKey);
        } catch ( IOException e) {
            throw new LoginException( "The MS-Logon failed, as the [" + excMsg + "] required by the DH key exchange could not be read.", e);
        }

        // Had to "do this by foot" because of http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6521495
        // DiffieHelman implementation coming with the JCE refuses to process the parameters generated by
        // the server as they are considered to be too short to generate a secure key.
        BigInteger biModulusX   = new BigInteger( zeMod);
        BigInteger biGeneratorX = new BigInteger( zeGen);
        BigInteger biKeyX       = new BigInteger( zeKey);
                
        //VNCViewer.pen.log( Level.FINEST, vc.logStr( "genX=" + biGeneratorX + ", modX=" + biModulusX));

        BigInteger prvKey = new BigInteger( biKeyX.bitLength() - 1, new SecureRandom());
        BigInteger pubKey = biGeneratorX.modPow( prvKey, biModulusX);

        try {
        	excMsg = "public key";
			os.write( longToBytes( pubKey.longValue()));
		} catch (IOException e1) {
            throw new LoginException( "The MS-Logon failed, as the [" + excMsg + "] required by the DH key exchange could not be sent.", e1);
		}

        BigInteger secKey = biKeyX.modPow( prvKey, biModulusX);
        secKey.shiftLeft( 1);
        byte secret[] =  longToBytes( secKey.longValue());

        if ( us.getBytes().length > 256)
        	VNCViewer.pen.log( Level.WARNING, vc.logStr( "The MS-Logon could fail as the username exceeds 256 bytes!"));

        if ( pw.getBytes().length > 64)
        	VNCViewer.pen.log( Level.WARNING, vc.logStr( "The MS-Logon could fail as the password exceeds 64 bytes!"));
        
        byte[] user   = pad( us.getBytes(), 256);
        byte[] passwd = pad( pw.getBytes(),  64);
        
        try {
        	excMsg = "username";
        	user   = encryptVNCDES( user,   secret, true, true);
        	excMsg = "password";
			passwd = encryptVNCDES( passwd, secret, true, true);
		} catch ( Exception e) {
            throw new LoginException( "The MS-Logon failed, as the [" + excMsg + "] could not be encrypted.", e);
		} 
        
        try {
            excMsg = "username";
            os.write( user);
            excMsg = "password";
            os.write( passwd);
        } catch ( IOException e) {
            throw new LoginException( "The MS-Logon failed, as the encrypted [" + excMsg + "] could not be sent.", e);
        }

        return isLoginSuccessful( "MS-Logon (DH)");
    }

    private byte[] pad( byte[] bIn, int len) {
    	if ( bIn == null) return new byte[ len];
    	if ( bIn.length >= len) return bIn;

    	byte[] rOutput = new byte[ len];
    	System.arraycopy( bIn, 0, rOutput, 0, bIn.length);
    	for ( int i = bIn.length; i < len; i++) rOutput[ i] = 0;
    	return rOutput;
    }

    private boolean isLoginSuccessful( String authTypeStr) throws LoginException {
        int i = 0;

		try {
		    VNCViewer.pen.log( Level.FINEST, vc.logStr( "Waiting 4 Server (Reading an Int) to confirm, it is happy to speak to me"));
			i = is.readInt();
            VNCViewer.pen.log( Level.FINEST, vc.logStr( "Got an answer from the server [" + i + "]"));
		} catch ( IOException e) {
			throw new LoginException( "Could not read the servers logon result message.", e);
		}

        Level l = Level.CONFIG;
        String connResMsg = null;
        AuthResult r = AuthResult.byValue( i);
        
        switch ( r) {
            case OK         :
            	l = Level.CONFIG;
            	connResMsg = "Authentification succeeded!";
            break;
            case FAILED     :
            	l = Level.WARNING;
            	readConnFailedReason( "Server refused the credentials.", r);
            break;
            case TOOMANY    :
            	l = Level.SEVERE;
				readConnFailedReason( "Server signalled being over capacity.", r);
            break;
            default			:
            	l = Level.SEVERE;
            	connResMsg = "Internal error: getMessage in VNCConnectionLA is missing a case in enum switch!:\n ";
        }
        
        VNCViewer.pen.log( l, vc.logStr( connResMsg));

        return r == RFB.AuthResult.OK;
    }

    private void readVersionMsg() throws LoginException {
        byte[] b = new byte[ 12];
        VNCViewer.pen.log( Level.FINEST, vc.logStr( "Reading version MSG"));

        if ( is == null) throw new LoginException( "Failed to read the servers version message as the input stream is NULL!");

        try {
			is.readFully( b);
		} catch ( IOException e) {
			throw new LoginException( "Could not read the servers version message.", e);
		}

        VNCViewer.pen.log( Level.FINEST, vc.logStr( "Server sent version MSG [" + ToolBox.escapeCRLF( new String( b)) + "]"));

        if (( b[  0] != 'R') || (b[  1] != 'F') || ( b[  2] != 'B') || ( b[  3] != ' ') ||
            ( b[  4] <  '0') || (b[  4] >  '9') || ( b[  5] <  '0') || ( b[  5] >  '9') ||
            ( b[  6] <  '0') || (b[  6] >  '9') || ( b[  7] != '.') || ( b[  8] <  '0') ||
            ( b[  8] >  '9') || (b[  9] <  '0') || ( b[  9] >  '9') || ( b[ 10] <  '0') ||
            ( b[ 10] >  '9') || (b[ 11] != '\n'))
        {
            throw new LoginException( "Host ["  + vc.getServerParams().getHostnameAndPort( false, false) + "] is not an RFB server");
        }

        int serverMajor = ( b[ 4] - '0') * 100 + ( b[ 5] - '0') * 10 + ( b[  6] - '0');
        int serverMinor = ( b[ 8] - '0') * 100 + ( b[ 9] - '0') * 10 + ( b[ 10] - '0');

        if(vc.getServerParams().isRemoteSupport()){
        	// SCIII supports only version RFB 3.3 ... but has a bug there so it sends RFB 3.16 to viewer - we have to ignore received version
        	serverMajor = 3;
        	serverMinor = 3;
        }
        
        vc.getServerParams().setVersion( true, serverMajor, serverMinor);

        if ( serverMajor < 3) throw new LoginException( "RFB server does not support protocol version 3");
    }

    private void writeVersionMsg() throws LoginException {
        vc.getServerParams().setVersionMajor( false, 3);
        String version2send = RFB.VERSIONMSG_3_8;

        if ( vc.getServerParams().getVersionMajor( true) > 3 ||
             vc.getServerParams().getVersionMinor( true) >= 8)
        {
            vc.getServerParams().setVersionMinor( false, 8);
        } else
            if ( vc.getServerParams().getVersionMinor( true) >= 7)
            {
                vc.getServerParams().setVersionMinor( false, 7);
                version2send = RFB.VERSIONMSG_3_7;
            } else {
                vc.getServerParams().setVersionMinor( false, 6);
                version2send = RFB.VERSIONMSG_3_6;
            }

        if(vc.getServerParams().isRemoteSupport()){
        	//SCIII supports only RFB 3.3
        	version2send = RFB.VERSIONMSG_3_3;
        }

        VNCViewer.pen.log( Level.FINEST, vc.logStr( "Sending version MSG [" + ToolBox.escapeCRLF( version2send) + "]"));
        
        try {
			os.write( version2send.getBytes());
			os.flush(); // added for SCIII - should be OK also with RC, if not FIXME
		} catch ( IOException e) {
			throw new LoginException( "Failed to send the desired protocol version (" + ToolBox.escapeCRLF( version2send) + ") to the server.", e);
		}
    }

    
    private void readConnFailedReason( String msg, AuthResult r) throws LoginException {
        try {
            int reasonLen = is.readInt();
            VNCViewer.pen.log( Level.FINEST, vc.logStr( "Read conn failed reason length: [" + reasonLen + "]"));
            byte[] reason = new byte[ reasonLen];
            is.readFully( reason);
            VNCViewer.pen.log( Level.FINEST, vc.logStr( "Read the conn failed reason"));

            throw new LoginException( r, msg, new String( reason, vc.getPresumedEncoding()));            
        } catch ( java.io.IOException e) {
        	String err = "Couldn't read  the connection failure reason.";
            VNCViewer.pen.log( Level.SEVERE, vc.logStr( err));
            throw new LoginException( err, e);            
        }
    }    
}
