package com.wincornixdorf.vnc.rfb.conf;

import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.Vector;

import com.wincornixdorf.vnc.rfb.debug.ToolBox;

/**This class mainly stores information on how to connect to a particular host.
 * Additionally it is able to store a "log path", which is used only by the infamous
 * log server, which requires a URL to connect to. The path is but a segment of said
 * URL and should best be ignored. */
public class LoginParameters {

	private String              originalHostname            = "";
    private String          	username    				= "";
    
    // TODO: store the password as a byte array and clear it after use!
    private String          	password    				= "";

    private InetSocketAddress   hostAddr    				= null;
	private boolean				gotInitializedByHostname	= true;

    private boolean				supportFailover				= false;

    /**The path variable is part of a URL the application will log to.*/
    private String          	path    					= "";

    private String sessionId = "";

    private boolean isRemoteSupport = false;

    public LoginParameters(boolean isRemoteSupport) {
        this.isRemoteSupport = isRemoteSupport;
    }

    public LoginParameters( boolean supportFailover, String hostName, int hostPort, String userName, String passwrd, boolean isRemoteSupport) {
        this.isRemoteSupport = isRemoteSupport;
        setHost( hostName, hostPort, false);
        this.username    		= userName;
        this.password    		= passwrd;
        this.supportFailover 	= supportFailover;
        this.sessionId = hostPort + "";
    }

    /**This constructor is geared towards the infamous log server. May it burn in HELL!
     * @param supportFailover I intend to kill this SHIT anyways.
     * @param logHostName I intend to kill this SHIT anyways.
     * @param userName I intend to kill this SHIT anyways.
     * @param passwrd I intend to kill this SHIT anyways.
     * @param logPath I intend to kill this SHIT anyways.*/
    public LoginParameters( boolean supportFailover, String logHostName, String userName, String passwrd, String logPath, boolean isRemoteSupport) {
        // TODO: WN-specific: The Portlet sends to a URL, NOT a hostname!!! (i.e.: https://foo.wincor-nixdorf.com:<port>
        this.isRemoteSupport = isRemoteSupport;
        setHost( logHostName, 80, true);
        this.username			= userName;
        this.password    		= passwrd;
        this.path				= logPath;
        this.supportFailover	= supportFailover;

    }
    
    public boolean isInitialized() { return hostAddr != null; }

    public boolean wasInitializedByHostname() { return gotInitializedByHostname; }

    public boolean hasAlternativeInetSocketAdresses() {
    	if ( !supportFailover) return false;
    	Vector<InetSocketAddress> v = getPotentialInetAdresses();
		return v.size() > 1;
    }

    public Vector<InetSocketAddress> getPotentialInetAdresses() {
    	Vector<InetSocketAddress> rval = new Vector<InetSocketAddress>();

    	if ( hostAddr == null) return rval;
    	
    	if ( !gotInitializedByHostname || !supportFailover) {
    		rval.add( hostAddr);
    		return rval;
    	}

		try {
			InetAddress[] candidates = InetAddress.getAllByName( originalHostname);
			
			for ( int i = 0; i < candidates.length; i++)
				rval.add( new InetSocketAddress( candidates[ i], hostAddr.getPort()));
		} catch ( UnknownHostException e) { }
		  catch ( SecurityException e) { }
		
		return rval;
    }
    
    public boolean hasAUsername() {
        return username != null && !username.isEmpty();
    }

    public int getPort() {
    	return hostAddr != null ? hostAddr.getPort() : -1;
    }

    public String getHostSpecial() {
    	InetAddress ia = hostAddr != null ? hostAddr.getAddress() : null;
    	boolean failedToResolveHostAddress = ia == null;
    	if ( failedToResolveHostAddress || !gotInitializedByHostname) return originalHostname;
    	return supportFailover ? hostAddr.getHostName() : hostAddr.getAddress().getHostAddress();
    }

    /**@return The name of the host as stated to the constructor
     * or an empty string.*/
    public String getOriginalHostname() { return originalHostname; }

    public String getUsername() { return username; }
    
    public String getPassword() { return password; }

    public InetSocketAddress getHostAddress() { return hostAddr; }
    
    public String getLogPath() { return path; }

    public URL getLogURL() {
    	try {
    		// TODO: Turns out this was not half bad: Remember the LOGHOST is in fact the stub of a URL!!!
    		// Intentionally using the original host name here, so the openConnection()
    		// method in URL will do any failover processing if the host name should
    		// map to multiple IP addresses.
			return new URL( originalHostname + path);
		} catch ( MalformedURLException e) {
			return null;
		}
    }

    private void setHost( String hostname, int port, boolean log) {
        if ( hostname == null || hostname.isEmpty()) return;

        gotInitializedByHostname	= !ToolBox.looksLikeIP4Address( hostname);

        if (isRemoteSupport && !log) {
            //For RS is port not important - set it to 1, because there usually comes high value, which is out of range
            hostAddr					= new InetSocketAddress( hostname, 1);
            //hostname contains 'ID' -> set it with port number
            originalHostname			= hostname + ":" + port;
        } else {
            hostAddr					= new InetSocketAddress( hostname, port);
            originalHostname			= hostname;
        }
    }

    public String getSessionId() {
        return sessionId;
    }
}
