/**
 * 
 */
package com.wincornixdorf.vnc.rfb.debug;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.Date;
import java.util.Observable;
import java.util.Observer;

import com.wincornixdorf.vnc.rfb.RFBHandler;
import com.wincornixdorf.vnc.rfb.RFBHandlerEvent;

/**
 * @author dic
 *
 * This class is designed to create high level debug tracefiles of VNC connections
 * by observing @link RFBHandlerEvent being emitted by @link RFBHandler instances.
 * Its sole purpose is to help debugging the protocol. 
 */
public class RFBHandlerObserver implements Observer {

    private static String   warn    = "Warning: ";
    
    private String          name    = "UNNAMED";
    private PrintStream     ps      = null;

    public RFBHandlerObserver( File f, String nameIn) {
        name = nameIn;

        try {
            ps = new PrintStream( f);
        } catch ( FileNotFoundException e) {
            System.err.println( warn + "Unable to observe RFBHandler [" + name + "]");
            e.printStackTrace();
            return;
        }
        
        System.err.println( "RFBHandlerObserver [" + name + "] will report to [" + f.toString() + "]");
        ps.println( "Trace date: [" + new Date() + "]");
        ps.println( "================================");
    }

    /* (non-Javadoc)
     * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
     */
    @Override
    public void update( Observable o, Object arg) {
        if ( !( o instanceof RFBHandler)) {
            System.err.println( warn + "RFBHandlerObserver received unsolicited event!");
            return;
        }
        
        if ( !( arg instanceof RFBHandlerEvent)) {
            System.err.println( warn + "RFBHandlerObserver received unknown event from RFBHandler!");
            return;
        }

        //RFBHandler      h = (RFBHandler)o;
        RFBHandlerEvent e = (RFBHandlerEvent)arg;
        
        e.dump( ps);
    }

}
