package com.wincornixdorf.vnc.rfb.decoder;

import java.awt.Color;
import java.io.DataInputStream;
import java.io.IOException;
import java.util.zip.Inflater;

import com.wincornixdorf.vnc.rfb.RFB;
import com.wincornixdorf.vnc.rfb.fb.RemoteDesktop;
import com.wincornixdorf.vnc.rfb.fb.VNCEncoding;

public class FBDecoderTight extends FBDecoder /*implements ImageObserver*/ {
    // Tight encoder's data.
	private final static int ZLIB_DATA__BUFFER_SIZE	= 8192;    
    private byte[]		buffer_zlibData		= new byte[ ZLIB_DATA__BUFFER_SIZE];

    private byte[]		palette08 			= new byte[ 2];
    private int[]		palette24 			= new int[ 256];
    
    private Inflater[]	tightInflaters		= new Inflater[ 4];

    public FBDecoderTight( RemoteDesktop rd) {
        super( rd);
        for ( int i = 0; i < 4; i++)
        	tightInflaters[ i] = new Inflater();
    }

    public VNCEncoding getEncoding() {
        return VNCEncoding.Tight;
    }

    public void process( DataInputStream is, int x, int y, int w, int h) throws Exception {
    	int bpp = rd.getPixelFormat().getDepth() / 8;
        
        // TODO: This line is so bogus, it makes a 13.57€ bill seem totally legit
    	//      int bpp = rd.getBytesPerPixel();
    	//        if ( bpp == 4) bpp = 3;

        int comp_ctl = is.readUnsignedByte();

        //if (rfb.rec != null) {
        //    if ( rfb.recordFromBeginning ||	comp_ctl == (rfb.TightFill << 4) ||
        //                                    comp_ctl == (rfb.TightJpeg << 4))
        //    {
        //        // Send data exactly as received.
        //        rfb.rec.writeByte(comp_ctl);
        //    } else {
        //        // Tell the decoder to flush each of the four zlib streams.
        //        rfb.rec.writeByte(comp_ctl | 0x0F);
        //    }
        //}

        // Flush zlib streams if we are told by the server to do so.
        for ( int stream_id = 0; stream_id < 4; stream_id++) {
            if (( comp_ctl & 1) != 0) tightInflaters[ stream_id] = new Inflater();
            comp_ctl >>= 1;
        }

        //System.out.println( "TR x[" + x + "] y[" + y + "] w[" + w + "] h[" + h + "] @ [" + rd.getBytesPerPixel() + "] comp_ctl [" + comp_ctl + "]");

        // Check correctness of subencoding value.
        if ( comp_ctl > RFB.TightMaxSubencoding)
            throw new Exception( "Incorrect tight subencoding: " + comp_ctl);

        // Handle solid-color rectangles.
        if ( comp_ctl == RFB.TightFill) {
            byte[] buf = new byte[ bpp];
            is.readFully( buf, 0, buf.length);

            switch ( bpp) {
                case 1  :
                	Color c = new Color( rd.cm08.getRGB( buf[ 0] & 0xFF));
                	rd.fillRect24( x, y, w, h, c);

                    // if (rfb.rec != null)
                    //  rfb.rec.writeByte(idx);
                break;
                case 2  :
                    rd.fillRect24( x, y, w, h, decode16BitColor( buf, 0));
                break;
                default:
                    // if (rfb.rec != null)
                    //  rfb.rec.write(buf);
                    rd.fillRect24( x, y, w, h, decode16BitColor5( buf));
            }
            rd.srcRDesktop.newPixels( x, y, w, h);
            return;
        }

        if ( comp_ctl == RFB.TightJpeg) {
            // Read JPEG data.
            byte[] jpegData = new byte[ readCompactLen( is)];
            is.readFully( jpegData, 0, jpegData.length);

            rd.writeJPEGData( jpegData, x, y, w, h);

            //if (rfb.rec != null) {
            //    if (!rfb.recordFromBeginning)
            //        rfb.recordCompactLen(jpegData.length);
            //    rfb.rec.write(jpegData);
            //}
            rd.srcRDesktop.newPixels( x, y, w, h);
            return;
        }

        // Read filter id and parameters.
        int numColors = 0, rowSize = w;
        boolean useGradient = false;

        if (( comp_ctl & RFB.TightExplicitFilter) != 0) {
            int filter_id = is.readUnsignedByte();

            // if (rfb.rec != null)
            //  rfb.rec.writeByte(filter_id);

            if ( filter_id == RFB.TightFilterPalette) {
                numColors = is.readUnsignedByte() + 1;

                byte[] buf = new byte[ numColors * bpp];

                switch ( bpp) {
                    case 1  :
                        if ( numColors != 2) throw new Exception( "Incorrect tight palette size: " + numColors);
                        is.readFully( palette08, 0, palette08.length);
                        //  if (rfb.rec != null)
                        //      rfb.rec.write(palette08);
                    break;
                    case 2  :
                        is.readFully( buf, 0, buf.length);
                        for ( int i = 0; i < numColors; i++)
                            palette24[ i] = decode16BitColor( buf, i * bpp).getRGB();
                    break;
                    default:
                        //if (rfb.rec != null)
                        //  rfb.rec.write(buf);
                        is.readFully( buf, 0, buf.length);
                        for ( int i = 0; i < numColors; i++)
                            palette24[ i] = decodeColor2( buf, i * bpp, bpp).getRGB();
                }

                if ( numColors == 2) rowSize = ( w + 7) / 8;

            } else if ( filter_id == RFB.TightFilterGradient) {
                useGradient = true;
            } else if ( filter_id != RFB.TightFilterCopy) {
                throw new Exception( "Incorrect tight filter id: " + filter_id);
            }
        }

        if ( numColors == 0) rowSize *= bpp;

        // Read, optionally uncompress and decode data.
        int dataSize = h * rowSize;
        //System.out.println( "rowsize [" + rowSize + "] datasize[" + dataSize + "]");
        if ( dataSize < RFB.TightMinToCompress) {
            // Data size is small - not compressed with zlib.
            //System.out.println( "Data size is small - numColors  [" + numColors + "]");

            if ( numColors != 0) {
                // Indexed colors.
                byte[] indexedData = new byte[ dataSize];
                is.readFully( indexedData, 0, indexedData.length);

                //if (rfb.rec != null)
                // rfb.rec.write(indexedData);

                if ( numColors == 2) {
                    // Two colors.
                    if ( rd.getBytesPerPixel() == 1)
                        decodeMonoData( x, y, w, h, indexedData, palette08);
                    else
                        decodeMonoData( x, y, w, h, indexedData, palette24);
                } else {
                    // 3..255 colors (assuming bytesPixel == 4).
                    int i = 0;
                    for ( int dy = y; dy < y + h; dy++) {
                        for ( int dx = x; dx < x + w; dx++)
                            rd.pixels24[ dy * rd.getFBW() + dx] = palette24[ indexedData[ i++] & 0xFF];
                    }
                }
            } else if ( useGradient) {
                // "Gradient"-processed data
                byte[] buf = new byte[ w * h * bpp];
                is.readFully( buf, 0, buf.length);

                // if (rfb.rec != null)
                //  rfb.rec.write(buf);

                decodeGradientData( x, y, w, h, buf);
            } else {
                // Raw truecolor data.
                processRAW( is, x, y, w, h, bpp);
            }
        } else {
            // Data was compressed with zlib.
            int zlibDataLen = readCompactLen( is);
            
            if ( buffer_zlibData.length < zlibDataLen)
            	buffer_zlibData = new byte[ zlibDataLen];
            
            is.readFully( buffer_zlibData, 0, zlibDataLen);

            // if (rfb.rec != null && rfb.recordFromBeginning)
            //  rfb.rec.write(zlibData);

            int stream_id = comp_ctl & 0x03;

            Inflater myInflater = tightInflaters[ stream_id];
            myInflater.setInput( buffer_zlibData, 0, zlibDataLen);
            byte[] buf = new byte[ dataSize];
            myInflater.inflate( buf);

            //if (rfb.rec != null && !rfb.recordFromBeginning)
            // rfb.recordCompressedData(buf);

            if ( numColors != 0) {
                // Indexed colors.
                if ( numColors == 2) {
                    // Two colors.
                    if ( rd.getBytesPerPixel() == 1)
                        decodeMonoData( x, y, w, h, buf, palette08);
                    else
                        decodeMonoData( x, y, w, h, buf, palette24);
                } else {
                    // More than two colors (assuming bytesPixel == 4).
                    int i = 0;
                    for ( int dy = y; dy < y + h; dy++) {
                        for ( int dx = x; dx < x + w; dx++)
                            rd.pixels24[ dy * rd.getFBW() + dx] = palette24[ buf[ i++] & 0xFF];
                    }
                }
            } else if ( useGradient) {
                // Compressed "Gradient"-filtered data (assuming bytesPixel == 4).
                    if ( rd.getBytesPerPixel() == 2) {
                        int i = -2;
                        int offset = y * rd.getFBW() + x;
                        for ( int dy = 0; dy < h; dy++) {
                            for ( int dx = 0; dx < w; dx++)
                                rd.pixels24[ offset++] = decode16BitColor( buf, i+=2).getRGB();
                            offset += ( rd.getFBW() - w);
                        }
                    } else
                        decodeGradientData( x, y, w, h, buf);
                } else {
                    int srcOffset = 0;
                    int destOffset;

                    // Compressed truecolor data.
                    switch ( bpp) {
                        case 1  :
                            destOffset = y * rd.getFBW() + x;
                            for ( int dy = 0; dy < h; dy++) {
                            	
                                //System.arraycopy( buf, dy * w, rd.pixels08, destOffset, w);

                            	for ( int i = 0; i < w; i++) {
                            		byte b = buf[ dy * w + i];
                            		Color c = new Color( rd.cm08.getRGB( b));
                                    rd.pixels24[ i + destOffset] = c.getRGB();
                            	}

                                destOffset += rd.getFBW();
                            }
                        break;
                        case 2  :
                            for ( int dy = 0; dy < h; dy++) {
                                myInflater.inflate( buf);
                                destOffset = ( y + dy) * rd.getFBW() + x;
                                for ( int i = 0; i < w; i++) {
                                    rd.pixels24[ destOffset + i] = decode16BitColor( buf, srcOffset).getRGB();
                                    srcOffset += bpp;
                                }
                            }
                        break;
                        default:
                            for ( int dy = 0; dy < h; dy++) {
                                myInflater.inflate( buf);
                                destOffset = ( y + dy) * rd.getFBW() + x;
                                for ( int i = 0; i < w; i++) {
                                    rd.pixels24[ destOffset + i] = decodeColor2( buf, srcOffset, bpp).getRGB();
                                    srcOffset += bpp;
                                }
                            }
                    }
                }
            }
        rd.srcRDesktop.newPixels( x, y, w, h);
    }

    // Decode 1bpp-encoded bi-color rectangle (8-bit and 24-bit versions).
    private void decodeMonoData( int x, int y, int w, int h, byte[] src, byte[] palette) {
        int dx, dy, n;
        int i = y * rd.getFBW() + x;
        int rowBytes = ( w + 7) / 8;

        for ( dy = 0; dy < h; dy++) {
            for ( dx = 0; dx < w / 8; dx++) {
                for ( n = 7; n >= 0; n--)
                	rd.pixels24[ i++] = palette[ src[ dy * rowBytes + dx] >> n & 1];
            }

            for ( n = 7; n >= 8 - w % 8; n--)
            	rd.pixels24[ i++] = palette[ src[ dy * rowBytes + dx] >> n & 1];

            i += ( rd.getFBW() - w);
        }
    }

    private void decodeMonoData( int x, int y, int w, int h, byte[] src, int[] palette) {
        int dx, dy, n;
        int i = y * rd.getFBW() + x;
        int rowBytes = ( w + 7) / 8;

        for ( dy = 0; dy < h; dy++) {
            for ( dx = 0; dx < w / 8; dx++) {
                for ( n = 7; n >= 0; n--)
                    rd.pixels24[ i++] = palette[ src[ dy * rowBytes + dx] >> n & 1];
            }

            for ( n = 7; n >= 8 - w % 8; n--)
                rd.pixels24[ i++] = palette[ src[ dy * rowBytes + dx] >> n & 1];

            i += ( rd.getFBW() - w);
        }
    }

    private void decodeGradientData( int x, int y, int w, int h, byte[] buf) {
        int bpp = rd.getBytesPerPixel();

        int dx, dy, c;
        byte[]  prevRow = new byte[ w * bpp];
        byte[]  thisRow = new byte[ w * bpp];
        byte[]  pix     = new byte[ bpp];
        int[]   est     = new int[ bpp];

        int offset = y * rd.getFBW() + x;

        for ( dy = 0; dy < h; dy++) {
            // First pixel in a row
            for ( c = 0; c < bpp; c++) {
                pix[ c] = (byte)( prevRow[ c] + buf[ dy * w * bpp + c]);
                thisRow[ c] = pix[ c];
            }

            rd.pixels24[ offset++] = decodeColor( pix, 0, 4).getRGB();

            // Remaining pixels of a row
            for ( dx = 1; dx < w; dx++) {
                for ( c = 0; c < bpp; c++) {
                    est[ c] = (( prevRow[   dx      * bpp + c] & 0xFF) + ( pix[ c] & 0xFF) -
                               ( prevRow[ ( dx - 1) * bpp + c] & 0xFF));

                    if ( est[ c] > 0xFF) est[ c] = 0xFF;
                    else if ( est[ c] < 0x00) est[ c] = 0x00;

                    pix[ c] = (byte)( est[ c] + buf[ ( dy * w + dx) * bpp + c]);
                    thisRow[ dx * bpp + c] = pix[ c];
                }

                rd.pixels24[ offset++] = decodeColor( pix, 0, 4).getRGB();
            }

            System.arraycopy( thisRow, 0, prevRow, 0, w * bpp);
            offset += ( rd.getFBW() - w);
        }
    }

    /**Read an integer in compact representation (1..3 bytes).
    Such format is used as a part of the Tight encoding.
    Also, this method records data if session recording is active and
    the viewer's recordFromBeginning variable is set to true.*/
    private int readCompactLen( DataInputStream is) throws IOException {
        //int     byteCount   = 1;
        int     len         = 0;
        int[]   portion     = new int[ 3];

        portion[ 0] = is.readUnsignedByte();
        len = portion[ 0] & 0x7F;

        if (( portion[ 0] & 0x80) != 0) {
            portion[ 1] = is.readUnsignedByte();
            //byteCount++;
            len |= ( portion[ 1] & 0x7F) << 7;
            if (( portion[ 1] & 0x80) != 0) {
                portion[ 2] = is.readUnsignedByte();
                //byteCount++;
                len |= ( portion[ 2] & 0xFF) << 14;
            }
        }

        //if ( rec != null && recordFromBeginning)
        //    for ( int i = 0; i < byteCount; i++)
        //        rec.writeByte( portion[ i]);

        return len;
    }
}
