package com.wincornixdorf.vnc.rfb.fb;

import java.io.IOException;
import java.io.OutputStream;
import java.io.DataInputStream;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Color;
import java.util.Observable;
import java.util.HashMap;
import java.util.logging.Level;
import java.awt.image.ColorModel;
import java.awt.image.MemoryImageSource;

import com.wincornixdorf.vnc.global.VNCViewer;
import com.wincornixdorf.vnc.rfb.RFB;
import com.wincornixdorf.vnc.rfb.RFBHandler;
import com.wincornixdorf.vnc.rfb.RFBHandlerEvent;
import com.wincornixdorf.vnc.rfb.decoder.FBDecoder;
import com.wincornixdorf.vnc.rfb.decoder.FBDecoderCoRRE;
import com.wincornixdorf.vnc.rfb.decoder.FBDecoderCopyRect;
import com.wincornixdorf.vnc.rfb.decoder.FBDecoderHextile;
import com.wincornixdorf.vnc.rfb.decoder.FBDecoderRAW;
import com.wincornixdorf.vnc.rfb.decoder.FBDecoderRRE;
import com.wincornixdorf.vnc.rfb.decoder.FBDecoderTight;
import com.wincornixdorf.vnc.rfb.decoder.FBDecoderZLib;
import com.wincornixdorf.vnc.rfb.decoder.FBDecoderZRLE;

public class RFBHandlerFB extends Observable {
    // These are SO PRIVATE! Hands off or my wrath shall be eternal!
    /**
     * This is a COPY of the sacred stream the session uses to read from the socket.
     * Come heaven, hell high waters or a newsflash saying hell had just frozen over
     * DO NOT SHARE THIS STREAM WITH ANY OTHER THREAD!
     */
    private DataInputStream is = null;
    /**
     * This is a COPY of the sacred stream the session uses to write to the socket.
     * Come heaven, hell high waters or a newsflash saying hell had just frozen over
     * DO NOT SHARE THIS STREAM WITH ANY OTHER THREAD!
     */
    private OutputStream os = null;

    private RemoteDesktop rd = new RemoteDesktop();

    /**
     * The decoder presently responsible for reading and extracting framebuffer packets.
     */
    private FBDecoder dec = null;

    /**
     * Number of FB rectangles pending to be read from the input stream.
     */
    private int updateNRects = -1;

    /**
     * X coordinate of the FB rectangle presently being processed.
     */
    private int updateRectX = -1;
    /**
     * Y coordinate of the FB rectangle presently being processed.
     */
    private int updateRectY = -1;
    /**
     * Width of the FB rectangle presently being processed.
     */
    private int updateRectW = -1;
    /**
     * Height of the FB rectangle presently being processed.
     */
    private int updateRectH = -1;
    /**
     * The encoding of the FB rectangle presently being processed.
     */
    private VNCEncoding updateRectEncoding = VNCEncoding.Auto;

    private boolean hadZlibUpdates = false;

    /**
     * If set true, the next FBUpdate request will demand a full update of the
     * Framebuffer. Not just an incremental one. Use with care!
     */
    private boolean fullUpdateNeeded = false;

    private boolean decoderChanged = false;

    /**
     * Used 4 internal reasons only. Will probably be removed soon. Do NOT use!
     */
    private RFBHandler vc = null;

    /**
     * A map for quickly retrieving previously used FBDecoders. This cache is
     * necessary as some encodings send packets in a different encoding than they
     * advertise. The createDecoder method would be forced into creating a new
     * one each time that happens, putting a heavy burden on the GC. Now once a
     * FBDecoder was produced, it is stored in this map and retrieved for reuse.
     */
    private HashMap<VNCEncoding, FBDecoder> hmDec = new HashMap<VNCEncoding, FBDecoder>();
    ;

    public RFBHandlerFB(RFBHandler c) {
        this.vc = c;

        // make sure we do at least have a Raw decoder in place
        updateRectEncoding = VNCEncoding.Raw;
        createDecoder();
    }

    //TODO: Storing references to the streams here is one VERY bad idea.
    public void setStreams(DataInputStream isOfSession, OutputStream osOfSession) {
//      if ( is == null) this.is = isOfSession;
//      if ( os == null) this.os = osOfSession;
        this.is = isOfSession;
        this.os = osOfSession;
    }

    public void handleFBUpdate() throws Exception {
        is.readByte(); // Padding
        updateNRects = is.readUnsignedShort();

        //VncViewer.pen.log( Level.FINEST, vc.logStr( "FramebufferUpdate message received."));
        //if ( updateNRects != 0) VncViewer.pen.log( Level.FINEST, vc.logStr( "Frames pending: " + updateNRects));

        for (int i = 0; i < updateNRects; i++) {
            onReceivingAFramebufferUpdateRectHdr();
            int rx = updateRectX, ry = updateRectY;
            int rw = updateRectW, rh = updateRectH;

            if (updateRectEncoding == VNCEncoding.LastRect)
                break;

            if (updateRectEncoding == VNCEncoding.NewFBSize) {
                VNCViewer.pen.log(Level.FINEST, vc.logStr("New Framebuffer Size!"));
                Dimension d = new Dimension(rw, rh);
                rd.setFBDim(d, false);
                onFBSizeChange(d);
                // updateFramebufferSize();
                break;
            }

            if (updateRectEncoding == VNCEncoding.XCursor ||
                    updateRectEncoding == VNCEncoding.RichCursor) {
                //VNCViewer.pen.log( Level.FINEST, vc.logStr( "New Cursor shape received!"));
                //int bpp = updateRectEncoding == VNCEncoding.XCursor ? rd.getBytesPerPixel() : 1;
                // Important! readNewCursorShape does read from the socket!!!
//            	int bytesPP = rd.getPixelFormat().getDepth() / 8;
                int bytesPP = rd.getBytesPerPixel();
//            	ColorModel cm = bpp == 1 ? rd.cm08 : rd.cm24;
                ColorModel cm = null;
                switch (bytesPP) {
                    case 1:
                        cm = rd.cm08;
                        break;
                    case 2:
                        cm = rd.cm8_64c;
                        break;
                    default:
                        cm = rd.cm24;
                        break;
                }
                rd.remoteCursor.readNewCursorShape(is, cm, updateRectEncoding, rx, ry, rw, rh, bytesPP);
                vc.fireEvent(RFBHandlerEvent.pdchanged(vc, rw, rh, true), null);
                continue;
            }

            if (updateRectEncoding == VNCEncoding.PointerPos) {
                rd.remoteCursor.setRemotePos(new Point(rx, ry));

                vc.fireEvent(RFBHandlerEvent.cursormoved(vc, rx, ry), null);
                //VNCViewer.pen.log( Level.FINEST, vc.logStr( "Cursor got moved!"));
                // softCursorMove(rx, ry);
                //cursorPosReceived = true;
                continue;
            }

            // startTiming();
            dec.process(is, rx, ry, rw, rh);
            // stopTiming();
        }

        // // Start/stop session recording if necessary. Request full
        // // update if a new session file was opened.
        // if ( viewer.checkRecordingStatus()) fullUpdateNeeded = true;

        requestNewPixelFormatWhereNecessary();

        writeFramebufferUpdateRequest(0, 0, getFBW(), getFBH(), !fullUpdateNeeded);
    }

    public void setRDProperties(PixelFormat pfProposedByServer, Dimension d, String desktopName, boolean requiresFBMemoryAllocation) {
        rd.setDesktopName(desktopName);
        //TODO: rd.setFBDim( d, !requiresFBMemoryAllocation); We are wasting memory here. Not asking 4 ScreenBufferUpdates doesn't prevent the server from sending any.
        rd.setFBDim(d, false);
        VNCViewer.pen.log(Level.FINEST, vc.logStr("Server proposed:\n" + PixelFormat.print(pfProposedByServer)));
        // rd.setPixelFormat( pfProposedByServer); Turns out I can't properly process that format. Server proposes 4Bytes@ depth 32
        sendPixelFormatRequest(vc.getSettingsCN().getPixelFormatQuality().getFormat());
        onFBSizeChange(d);
    }

    private void requestNewPixelFormatWhereNecessary() {
        PixelFormat desiredFormat = vc.getSettingsCN().getPixelFormatQuality().getFormat();
        if (desiredFormat.equals(rd.getPixelFormat())) return;

        sendPixelFormatRequest(desiredFormat);
    }

    private void sendPixelFormatRequest(PixelFormat pf) {
        VNCViewer.pen.log(Level.FINEST, vc.logStr("Sending PF request:\n" + PixelFormat.print(pf)));

        try {
            os.write(pf.getAsRFBProtocolCompliantByteArray());
            rd.setPixelFormat(pf);

            // having requested a different pixel format the contents of
            // the Framebuffer are outdated and must be updated completely
            fullUpdateNeeded = true;

            onFBSizeChange(null);
        } catch (IOException e) {
            VNCViewer.pen.log(Level.WARNING, vc.logStr("Failed to negotiate the pixel format with the server."));
            e.printStackTrace();
        }
    }

    public void handleColorMapEntries() throws IOException {
        is.readByte(); // Padding
        int firstColor = is.readUnsignedShort();
        int numberOfColors = is.readUnsignedShort();
        int i = firstColor;

        //System.out.println( "CME firstCol:" + firstColor + " Cols#" + numberOfColors);

        while (i < numberOfColors && i < rd.fixedPalette.length) {
            int r = is.readUnsignedShort();
            int g = is.readUnsignedShort();
            int b = is.readUnsignedShort();
            //System.out.println( "r/g/b: [" + r + "/" + g + "/" + b + ")");
            rd.fixedPalette[i++] = new Color(r, g, b);
        }
    }

    public Dimension getFBDim() {
        return rd.getFBDim();
    }

    public int getFBW() {
        return rd.getFBW();
    }

    public int getFBH() {
        return rd.getFBH();
    }

    public int getBytesPerPixel() {
        return rd.getBytesPerPixel();
    }

    public MemoryImageSource getRemoteDesktopSource() {
        return rd.srcRDesktop;
    }

    public RemoteCursor getRemoteCursor() {
        return rd.remoteCursor;
    }

    public String getRemoteDesktopName() {
        return rd.getDesktopName();
    }

    public void requestInitialFB() throws IOException {
        VNCViewer.pen.log(Level.FINEST, "WTF_initial");
        writeFramebufferUpdateRequest(0, 0, getFBW(), getFBH(), true);
    }

    public void requestFullFBUpdate() throws IOException {
        VNCViewer.pen.log(Level.FINEST, "WTF_full");
        writeFramebufferUpdateRequest(0, 0, getFBW(), getFBH(), false);
    }

    public boolean hadZLibUpdates() {
        return hadZlibUpdates;
    }

    /**
     * Users can set the encoding using the fancy GUI all they want. If the
     * server does not offer the requested encoding it will fall back to whatever
     * encoding it does know. If all else fails, the server will probably start
     * sending RAW encoded images.
     *
     * @return The encoding actually used by the server to encode the FB data.
     */
    public VNCEncoding getEncodingUsedByServer() {
        return dec != null ? dec.getEncoding() : VNCEncoding.Auto;
    }

    /**
     * Will write a Framebuffer update request to the server. Call this method
     * only from the FB connection and never from the FT one.
     */
    private void writeFramebufferUpdateRequest(int x, int y, int w, int h, boolean incremental) throws IOException {
        if (!vc.canHandleFramebufferUpdates()) return;

        //String s = String.format( "FBUpdRequest [x|y|w|h|incremental]: [%1$4d|%2$4d|%3$4d|%4$4d|%5$8b]", x, y, w, h, incremental);
        //VNCViewer.pen.log( Level.FINEST, vc.logStr( s));

        byte[] b = new byte[10];

        b[0] = (byte) RFB.FramebufferUpdateRequest;
        b[1] = (byte) (incremental ? 1 : 0);
        b[2] = (byte) ((x >> 8) & 0xff);
        b[3] = (byte) (x & 0xff);
        b[4] = (byte) ((y >> 8) & 0xff);
        b[5] = (byte) (y & 0xff);
        b[6] = (byte) ((w >> 8) & 0xff);
        b[7] = (byte) (w & 0xff);
        b[8] = (byte) ((h >> 8) & 0xff);
        b[9] = (byte) (h & 0xff);

        // Make sure we don't spam the server with full update requests.
        fullUpdateNeeded = false;

        os.write(b);

    }

    private void onReceivingAFramebufferUpdateRectHdr() throws Exception {
        updateRectX = is.readUnsignedShort();
        updateRectY = is.readUnsignedShort();
        updateRectW = is.readUnsignedShort();
        updateRectH = is.readUnsignedShort();
        updateRectEncoding = VNCEncoding.getEncoding(is.readInt());

//        VNCViewer.pen.log( Level.FINEST, String.format( "FBTile       [x|y|w|h|encoding   ]: [%1$4d|%2$4d|%3$4d|%4$4d|%5$8s]", 
//        		updateRectX, updateRectY, updateRectW, updateRectH, updateRectEncoding.toString()));

        if (updateRectEncoding == VNCEncoding.Zlib ||
                updateRectEncoding == VNCEncoding.ZRLE ||
                updateRectEncoding == VNCEncoding.Tight)
            hadZlibUpdates = true;

        /*// If the session is being recorded:
        if (rec != null) {
            if (numUpdatesInSession > 1)
                rec.flush();		// Flush the output on each rectangle.
            rec.writeShortBE(updateRectX);
            rec.writeShortBE(updateRectY);
            rec.writeShortBE(updateRectW);
            rec.writeShortBE(updateRectH);
            if (updateRectEncoding == EncodingZlib && !recordFromBeginning) {
                // Here we cannot write Zlib-encoded rectangles because the
                // decoder won't be able to reproduce zlib stream state.
                if (!zlibWarningShown) {
                    System.out.println("Warning: Raw encoding will be used " +
                            "instead of Zlib in recorded session.");
                    zlibWarningShown = true;
                }

                rec.writeIntBE(EncodingRaw);
            } else {
                rec.writeIntBE(updateRectEncoding);
                if (updateRectEncoding == EncodingTight && !recordFromBeginning && !tightWarningShown) {
                    System.out.println("Warning: Re-compressing Tight-encoded " +
                        "updates for session recording.");
                    tightWarningShown = true;
                }
            }
        }*/

        if (updateRectEncoding.toInt() < 0 || updateRectEncoding.toInt() > RFB.MaxNormalEncoding) return;

        if (updateRectX + updateRectW > rd.getFBW() ||
                updateRectY + updateRectH > rd.getFBH()) {
            throw new Exception("Framebuffer update rectangle too large: " +
                    updateRectW + "x" + updateRectH + " at (" +
                    updateRectX + "," + updateRectY + ")");
        }

        createDecoder();
    }

    private void createDecoder() {
        String mode = "Reusing an existing Decoder object.";
        VNCEncoding oldEncoding = getEncodingUsedByServer();
        if (oldEncoding == updateRectEncoding)
            return;

        // Just to be on the safe side ....
        if (!updateRectEncoding.isSupported() || updateRectEncoding.isPseudo()) return;

        dec = hmDec.get(updateRectEncoding);

        if (dec == null) {
            mode = "Had to create a new Decoder object.";
            switch (updateRectEncoding.toInt()) {
                case RFB.EncodingTight:
                    hmDec.put(VNCEncoding.Tight, new FBDecoderTight(rd));
                    break;
                case RFB.EncodingRaw:
                    hmDec.put(VNCEncoding.Raw, new FBDecoderRAW(rd));
                    break;
                case RFB.EncodingCopyRect:
                    hmDec.put(VNCEncoding.CopyRect, new FBDecoderCopyRect(rd));
                    break;
                case RFB.EncodingRRE:
                    hmDec.put(VNCEncoding.RRE, new FBDecoderRRE(rd));
                    break;
                case RFB.EncodingCoRRE:
                    hmDec.put(VNCEncoding.CoRRE, new FBDecoderCoRRE(rd));
                    break;
                case RFB.EncodingHextile:
                    hmDec.put(VNCEncoding.Hextile, new FBDecoderHextile(rd));
                    break;
                case RFB.EncodingZRLE:
                    hmDec.put(VNCEncoding.ZRLE, new FBDecoderZRLE(rd));
                    break;
                case RFB.EncodingZlib:
                    hmDec.put(VNCEncoding.Zlib, new FBDecoderZLib(rd));
                    break;
                default:
                    hmDec.put(updateRectEncoding, new FBDecoderTight(rd));
                    break;
            }
            dec = hmDec.get(updateRectEncoding);
        }

        decoderChanged = true;
        String od = oldEncoding.toString();
        String nd = dec.getEncoding().toString();
        VNCViewer.pen.log(Level.FINEST, vc.logStr("Switched from a [" + od + "] to [" + nd + "] Decoder. " + mode));
    }

    /**
     * If a RFB Handler is to be reused, this method must 1st be invoked, as it
     * will make sure the decoders are in their initial position. Some decoders
     * are keeping runtime data of their own, which depends on the data stream
     * read so far. On trying to reuse a decoder with a different stream, these
     * decoders will crash as they will suddenly be confronted with data that
     * does not agree with what may have been read in the previous session.
     */
    public void resetDecoders() {
        dec = null;
        hmDec.clear();
    }

    private void onFBSizeChange(Dimension d) {
        int w = d != null ? d.width : -1;
        int h = d != null ? d.height : -1;
        vc.fireEvent(RFBHandlerEvent.pdchanged(vc, w, h, false), null);
    }

    public boolean isDecoderChanged() {
        return decoderChanged;
    }

    public void setDecoderChanged(boolean decoderChanged) {
        this.decoderChanged = decoderChanged;
    }

    /**
     * Resets decoder settings to the default values.
     */
    public void processReconnecting() {
        resetDecoders();
        updateRectEncoding = VNCEncoding.Raw;
        createDecoder();
    }
}
